import pytest

from pydantic_core import ArgsKwargs, ValidationError
from pydantic_core import core_schema as cs

from ...conftest import Err, PyAndJson


@pytest.mark.parametrize(
    ['input_value', 'expected'],
    (
        [ArgsKwargs((), {'x': 1}), ((), {'x': 1})],
        [ArgsKwargs((), {'x': 1, 'z': True}), ((), {'x': 1, 'y': True})],
        [ArgsKwargs((), {}), Err('', [{'type': 'missing', 'loc': ('x',)}])],
        [ArgsKwargs((), {'x': 'not_an_int'}), Err('', [{'type': 'int_parsing', 'loc': ('x',)}])],
        [ArgsKwargs((), {'x': 1, 'y': True}), Err('', [{'type': 'extra_forbidden', 'loc': ('y',)}])],
        [{'kwargs': {'x': 1}}, ((), {'x': 1})],
        [{'kwargs': {'x': 1, 'z': True}}, ((), {'x': 1, 'y': True})],
        [{'kwargs': {}}, Err('', [{'type': 'missing', 'loc': ('kwargs', 'x')}])],
        [{}, Err('', [{'type': 'missing', 'loc': ('kwargs', 'x')}])],
        [
            {'kwargs': {'x': 'not_an_int'}},
            Err(
                '',
                [
                    {
                        'type': 'int_parsing',
                        'loc': (
                            'kwargs',
                            'x',
                        ),
                    }
                ],
            ),
        ],
        [
            {'kwargs': {'x': 1, 'y': True}},
            Err(
                '',
                [
                    {
                        'type': 'extra_forbidden',
                        'loc': (
                            'kwargs',
                            'y',
                        ),
                    }
                ],
            ),
        ],
    ),
)
def test_var_kwargs(py_and_json: PyAndJson, input_value, expected) -> None:
    """Test (in)valid inputs against var-args parameters (unpacked typed dict):

    ```python
    class TD(TypedDict, total=false):
        x: Required[int]
        y: Annotated[bool, Field(validation_alias='z')]

    def func(**kwargs: Unpack[TD]):
        ...
    ```
    """
    v = py_and_json(
        cs.arguments_v3_schema(
            [
                cs.arguments_v3_parameter(
                    name='kwargs',
                    schema=cs.typed_dict_schema(
                        {
                            'x': cs.typed_dict_field(
                                schema=cs.int_schema(),
                                required=True,
                            ),
                            'y': cs.typed_dict_field(
                                schema=cs.bool_schema(),
                                required=False,
                                validation_alias='z',
                            ),
                        },
                        extra_behavior='forbid',
                    ),
                    mode='var_kwargs_unpacked_typed_dict',
                ),
            ]
        )
    )

    if isinstance(expected, Err):
        with pytest.raises(ValidationError) as exc_info:
            v.validate_test(input_value)

        error = exc_info.value.errors()[0]

        assert error['type'] == expected.errors[0]['type']
        assert error['loc'] == expected.errors[0]['loc']
    else:
        assert v.validate_test(input_value) == expected


def test_var_kwargs_invalid_dict(py_and_json: PyAndJson) -> None:
    """Test invalid dict-like input against var-kwargs parameters in mapping validation mode."""
    v = py_and_json(
        cs.arguments_v3_schema(
            [
                cs.arguments_v3_parameter(
                    name='kwargs', schema=cs.typed_dict_schema({}), mode='var_kwargs_unpacked_typed_dict'
                ),
            ]
        )
    )

    with pytest.raises(ValidationError) as exc_info:
        v.validate_test({'kwargs': 'not_a_dict'})

    error = exc_info.value.errors()[0]

    assert error['type'] == 'dict_type'
    assert error['loc'] == ('kwargs',)



import dataclasses
import json
import platform
import sys
from typing import ClassVar

import pytest

from pydantic_core import SchemaSerializer, SchemaValidator, core_schema

on_pypy = platform.python_implementation() == 'PyPy'
# pypy doesn't seem to maintain order of `__dict__`
if on_pypy:
    IsStrictDict = dict
else:
    from dirty_equals import IsStrictDict


@dataclasses.dataclass
class Foo:
    a: str
    b: bytes


def test_dataclass():
    schema = core_schema.dataclass_schema(
        Foo,
        core_schema.dataclass_args_schema(
            'Foo',
            [
                core_schema.dataclass_field(name='a', schema=core_schema.str_schema()),
                core_schema.dataclass_field(name='b', schema=core_schema.bytes_schema()),
            ],
        ),
        ['a', 'b'],
    )
    s = SchemaSerializer(schema)
    assert s.to_python(Foo(a='hello', b=b'more')) == IsStrictDict(a='hello', b=b'more')
    assert s.to_python(Foo(a='hello', b=b'more'), mode='json') == IsStrictDict(a='hello', b='more')
    j = s.to_json(Foo(a='hello', b=b'more'))

    if on_pypy:
        assert json.loads(j) == {'a': 'hello', 'b': 'more'}
    else:
        assert j == b'{"a":"hello","b":"more"}'

    assert s.to_python(Foo(a='hello', b=b'more'), exclude={'b'}) == IsStrictDict(a='hello')
    assert s.to_json(Foo(a='hello', b=b'more'), include={'a'}) == b'{"a":"hello"}'


def test_serialization_exclude():
    schema = core_schema.dataclass_schema(
        Foo,
        core_schema.dataclass_args_schema(
            'Foo',
            [
                core_schema.dataclass_field(
                    name='a', schema=core_schema.str_schema(), serialization_exclude_if=lambda x: x == 'bye'
                ),
                core_schema.dataclass_field(name='b', schema=core_schema.bytes_schema(), serialization_exclude=True),
            ],
        ),
        ['a', 'b'],
    )
    s = SchemaSerializer(schema)
    assert s.to_python(Foo(a='hello', b=b'more')) == {'a': 'hello'}
    assert s.to_python(Foo(a='hello', b=b'more'), mode='json') == {'a': 'hello'}
    # a = 'bye' excludes it
    assert s.to_python(Foo(a='bye', b=b'more'), mode='json') == {}
    j = s.to_json(Foo(a='hello', b=b'more'))
    if on_pypy:
        assert json.loads(j) == {'a': 'hello'}
    else:
        assert j == b'{"a":"hello"}'
    j = s.to_json(Foo(a='bye', b=b'more'))
    if on_pypy:
        assert json.loads(j) == {}
    else:
        assert j == b'{}'


def test_serialization_alias():
    schema = core_schema.dataclass_schema(
        Foo,
        core_schema.dataclass_args_schema(
            'Foo',
            [
                core_schema.dataclass_field(name='a', schema=core_schema.str_schema()),
                core_schema.dataclass_field(name='b', schema=core_schema.bytes_schema(), serialization_alias='BAR'),
            ],
        ),
        ['a', 'b'],
    )
    s = SchemaSerializer(schema)
    assert s.to_python(Foo(a='hello', b=b'more'), by_alias=True) == IsStrictDict(a='hello', BAR=b'more')
    assert s.to_python(Foo(a='hello', b=b'more'), mode='json', by_alias=True) == IsStrictDict(a='hello', BAR='more')
    j = s.to_json(Foo(a='hello', b=b'more'), by_alias=True)

    if on_pypy:
        assert json.loads(j) == {'a': 'hello', 'BAR': 'more'}
    else:
        assert j == b'{"a":"hello","BAR":"more"}'


def test_properties():
    @dataclasses.dataclass
    class FooProp:
        a: str
        b: bytes

        @property
        def c(self) -> str:
            return f'{self.a} {self.b.decode()}'

    schema = core_schema.dataclass_schema(
        Foo,
        core_schema.dataclass_args_schema(
            'FooProp',
            [
                core_schema.dataclass_field(name='a', schema=core_schema.str_schema()),
                core_schema.dataclass_field(name='b', schema=core_schema.bytes_schema()),
            ],
            computed_fields=[core_schema.computed_field('c', core_schema.str_schema())],
        ),
        ['a', 'b'],
    )
    s = SchemaSerializer(schema)
    assert s.to_python(FooProp(a='hello', b=b'more')) == IsStrictDict(a='hello', b=b'more', c='hello more')
    assert s.to_python(FooProp(a='hello', b=b'more'), mode='json') == IsStrictDict(a='hello', b='more', c='hello more')
    j = s.to_json(FooProp(a='hello', b=b'more'))

    if on_pypy:
        assert json.loads(j) == {'a': 'hello', 'b': 'more', 'c': 'hello more'}
    else:
        assert j == b'{"a":"hello","b":"more","c":"hello more"}'

    assert s.to_python(FooProp(a='hello', b=b'more'), exclude={'b'}) == IsStrictDict(a='hello', c='hello more')
    assert s.to_json(FooProp(a='hello', b=b'more'), include={'a'}) == b'{"a":"hello"}'


@pytest.mark.skipif(sys.version_info < (3, 10), reason='slots are only supported for dataclasses in Python > 3.10')
def test_slots_mixed():
    @dataclasses.dataclass(slots=True)
    class Model:
        x: int
        y: dataclasses.InitVar[str]
        z: ClassVar[str] = 'z-classvar'

    @dataclasses.dataclass
    class SubModel(Model):
        x2: int
        y2: dataclasses.InitVar[str]
        z2: ClassVar[str] = 'z2-classvar'

    schema = core_schema.dataclass_schema(
        SubModel,
        core_schema.dataclass_args_schema(
            'SubModel',
            [
                core_schema.dataclass_field(name='x', schema=core_schema.int_schema()),
                core_schema.dataclass_field(name='y', init_only=True, schema=core_schema.str_schema()),
                core_schema.dataclass_field(name='x2', schema=core_schema.int_schema()),
                core_schema.dataclass_field(name='y2', init_only=True, schema=core_schema.str_schema()),
            ],
        ),
        ['x', 'x2'],
        slots=True,
    )
    dc = SubModel(x=1, y='a', x2=2, y2='b')
    assert dataclasses.asdict(dc) == {'x': 1, 'x2': 2}

    s = SchemaSerializer(schema)
    assert s.to_python(dc) == {'x': 1, 'x2': 2}
    assert s.to_json(dc) == b'{"x":1,"x2":2}'


@pytest.mark.xfail(reason='dataclasses do not serialize extras')
def test_extra_custom_serializer():
    @dataclasses.dataclass
    class Model:
        pass

    schema = core_schema.dataclass_schema(
        Model,
        core_schema.dataclass_args_schema(
            'Model',
            [],
            extra_behavior='allow',
            # extras_schema=core_schema.any_schema(
            #     serialization=core_schema.plain_serializer_function_ser_schema(
            #         lambda v: v + ' bam!',
            #     )
            # )
        ),
        [],
    )
    s = SchemaSerializer(schema)
    v = SchemaValidator(schema)

    m = v.validate_python({'extra': 'extra'})

    assert s.to_python(m) == {'extra': 'extra bam!'}


def test_dataclass_initvar_not_required_on_union_ser() -> None:
    @dataclasses.dataclass
    class Foo:
        x: int
        init_var: dataclasses.InitVar[int] = 1

    @dataclasses.dataclass
    class Bar:
        x: int

    schema = core_schema.union_schema(
        [
            core_schema.dataclass_schema(
                Foo,
                core_schema.dataclass_args_schema(
                    'Foo',
                    [
                        core_schema.dataclass_field(name='x', schema=core_schema.int_schema()),
                        core_schema.dataclass_field(
                            name='init_var',
                            init_only=True,
                            schema=core_schema.with_default_schema(core_schema.int_schema(), default=1),
                        ),
                    ],
                ),
                ['x'],
                post_init=True,
            ),
            core_schema.dataclass_schema(
                Bar,
                core_schema.dataclass_args_schema(
                    'Bar', [core_schema.dataclass_field(name='x', schema=core_schema.int_schema())]
                ),
                ['x'],
            ),
        ]
    )

    s = SchemaSerializer(schema)
    assert s.to_python(Foo(x=1), warnings='error') == {'x': 1}
    assert s.to_python(Foo(x=1, init_var=2), warnings='error') == {'x': 1}


@pytest.mark.parametrize(
    'config,runtime,expected',
    [
        (True, True, {'my_alias': 'hello'}),
        (True, False, {'my_field': 'hello'}),
        (True, None, {'my_alias': 'hello'}),
        (False, True, {'my_alias': 'hello'}),
        (False, False, {'my_field': 'hello'}),
        (False, None, {'my_field': 'hello'}),
        (None, True, {'my_alias': 'hello'}),
        (None, False, {'my_field': 'hello'}),
        (None, None, {'my_field': 'hello'}),
    ],
)
def test_by_alias_and_name_config_interaction(config, runtime, expected) -> None:
    """This test reflects the priority that applies for config vs runtime serialization alias configuration.

    If the runtime value (by_alias) is set, that value is used.
    If the runtime value is unset, the config value (serialize_by_alias) is used.
    If neither are set, the default, False, is used.
    """

    @dataclasses.dataclass
    class Foo:
        my_field: str

    schema = core_schema.dataclass_schema(
        Foo,
        core_schema.dataclass_args_schema(
            'Foo',
            [
                core_schema.dataclass_field(
                    name='my_field', schema=core_schema.str_schema(), serialization_alias='my_alias'
                ),
            ],
        ),
        ['my_field'],
        config=core_schema.CoreConfig(serialize_by_alias=config or False),
    )
    s = SchemaSerializer(schema)
    assert s.to_python(Foo(my_field='hello'), by_alias=runtime) == expected



import pytest

from pydantic_core import ArgsKwargs, ValidationError
from pydantic_core import core_schema as cs

from ...conftest import PyAndJson


@pytest.mark.parametrize(
    'input_value',
    [
        ArgsKwargs((1, True)),
        ArgsKwargs((1,)),
        {'a': 1, 'b': True},
        {'a': 1},
    ],
)
def test_positional_only(py_and_json: PyAndJson, input_value) -> None:
    """Test valid inputs against positional-only parameters:

    ```python
    def func(a: int, b: bool = True, /):
        ...
    ```
    """
    v = py_and_json(
        cs.arguments_v3_schema(
            [
                cs.arguments_v3_parameter(name='a', schema=cs.int_schema(), mode='positional_only'),
                cs.arguments_v3_parameter(
                    name='b', schema=cs.with_default_schema(cs.bool_schema(), default=True), mode='positional_only'
                ),
            ]
        )
    )

    assert v.validate_test(input_value) == ((1, True), {})


def test_positional_only_validation_error(py_and_json: PyAndJson) -> None:
    """Test invalid inputs against positional-only parameters:

    ```python
    def func(a: int, /):
        ...

    func('not_an_int')
    ```
    """
    v = py_and_json(
        cs.arguments_v3_schema(
            [
                cs.arguments_v3_parameter(name='a', schema=cs.int_schema(), mode='positional_only'),
            ]
        )
    )

    with pytest.raises(ValidationError) as exc_info:
        v.validate_test(ArgsKwargs(('not_an_int',), {}))

    error = exc_info.value.errors()[0]

    assert error['type'] == 'int_parsing'
    assert error['loc'] == (0,)

    with pytest.raises(ValidationError) as exc_info:
        v.validate_test({'a': 'not_an_int'})

    error = exc_info.value.errors()[0]

    assert error['type'] == 'int_parsing'
    assert error['loc'] == ('a',)


def test_positional_only_error_required(py_and_json: PyAndJson) -> None:
    """Test missing inputs against positional-only parameters:

    ```python
    def func(a: int, /):
        ...

    func()
    ```
    """
    v = py_and_json(
        cs.arguments_v3_schema(
            [
                cs.arguments_v3_parameter(name='a', schema=cs.int_schema(), mode='positional_only'),
            ]
        )
    )

    with pytest.raises(ValidationError) as exc_info:
        v.validate_test(ArgsKwargs((), {}))

    error = exc_info.value.errors()[0]

    assert error['type'] == 'missing_positional_only_argument'
    assert error['loc'] == (0,)

    with pytest.raises(ValidationError) as exc_info:
        v.validate_test({})

    error = exc_info.value.errors()[0]

    assert error['type'] == 'missing_positional_only_argument'
    assert error['loc'] == ('a',)



from enum import Enum

from pydantic_core import SchemaSerializer, core_schema


# serializing enum calls methods in serializers::infer
def test_infer_to_python():
    class MyEnum(Enum):
        complex_ = complex(1, 2)

    v = SchemaSerializer(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values())))
    assert v.to_python(MyEnum.complex_, mode='json') == '1+2j'


def test_infer_serialize():
    class MyEnum(Enum):
        complex_ = complex(1, 2)

    v = SchemaSerializer(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values())))
    assert v.to_json(MyEnum.complex_) == b'"1+2j"'


def test_infer_json_key():
    class MyEnum(Enum):
        complex_ = {complex(1, 2): 1}

    v = SchemaSerializer(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values())))
    assert v.to_json(MyEnum.complex_) == b'{"1+2j":1}'



from datetime import date, datetime, time, timedelta, timezone

import pytest

from pydantic_core import SchemaSerializer, core_schema


def test_datetime():
    v = SchemaSerializer(core_schema.datetime_schema())
    assert v.to_python(datetime(2022, 12, 2, 12, 13, 14)) == datetime(2022, 12, 2, 12, 13, 14)

    assert v.to_python(datetime(2022, 12, 2, 12, 13, 14), mode='json') == '2022-12-02T12:13:14'
    assert v.to_json(datetime(2022, 12, 2, 12, 13, 14)) == b'"2022-12-02T12:13:14"'

    with pytest.warns(
        UserWarning,
        match=r'Expected `datetime` - serialized value may not be as expected \[input_value=123, input_type=int\]',
    ):
        assert v.to_python(123, mode='json') == 123

    with pytest.warns(
        UserWarning,
        match=r'Expected `datetime` - serialized value may not be as expected \[input_value=123, input_type=int\]',
    ):
        assert v.to_json(123) == b'123'


def test_datetime_key():
    v = SchemaSerializer(core_schema.dict_schema(core_schema.datetime_schema(), core_schema.datetime_schema()))
    assert v.to_python({datetime(2022, 12, 2, 12, 13, 14): datetime(2022, 12, 2, 12, 13, 14)}) == {
        datetime(2022, 12, 2, 12, 13, 14): datetime(2022, 12, 2, 12, 13, 14)
    }
    assert v.to_python({datetime(2022, 12, 2, 12, 13, 14): datetime(2022, 12, 2, 12, 13, 14)}, mode='json') == {
        '2022-12-02T12:13:14': '2022-12-02T12:13:14'
    }
    assert (
        v.to_json({datetime(2022, 12, 2, 12, 13, 14): datetime(2022, 12, 2, 12, 13, 14)})
        == b'{"2022-12-02T12:13:14":"2022-12-02T12:13:14"}'
    )


def tz(**kwargs):
    return timezone(timedelta(**kwargs))


@pytest.mark.parametrize(
    'value,expected',
    [
        (datetime(2022, 12, 2, 12, 13, 14), '2022-12-02T12:13:14'),
        (datetime(2022, 12, 2, 12, tzinfo=timezone.utc), '2022-12-02T12:00:00Z'),
        (datetime(2022, 12, 2, 12, tzinfo=tz(hours=2)), '2022-12-02T12:00:00+02:00'),
        (datetime(2022, 12, 2, 12, tzinfo=tz(hours=2, minutes=30)), '2022-12-02T12:00:00+02:30'),
        (datetime(2022, 12, 2, 12, tzinfo=tz(hours=-2)), '2022-12-02T12:00:00-02:00'),
        (datetime(2022, 12, 2, 12, tzinfo=tz(hours=-2, minutes=-30)), '2022-12-02T12:00:00-02:30'),
        (datetime(2022, 12, 2, 12, 13, 14, 123456), '2022-12-02T12:13:14.123456'),
        (datetime(2022, 12, 2, 12, 13, 14, 123), '2022-12-02T12:13:14.000123'),
        (datetime(2022, 12, 2, 12, 13, 14, 123_000), '2022-12-02T12:13:14.123000'),
        (datetime(2022, 12, 2, 12, 13, 14, 123456, tzinfo=tz(hours=-2)), '2022-12-02T12:13:14.123456-02:00'),
    ],
)
def test_datetime_json(value, expected):
    v = SchemaSerializer(core_schema.datetime_schema())
    assert v.to_python(value, mode='json') == expected
    assert v.to_json(value).decode() == f'"{expected}"'


def test_date():
    v = SchemaSerializer(core_schema.date_schema())
    assert v.to_python(date(2022, 12, 2)) == date(2022, 12, 2)

    assert v.to_python(date(2022, 12, 2), mode='json') == '2022-12-02'
    assert v.to_json(date(2022, 12, 2)) == b'"2022-12-02"'


def test_date_key():
    v = SchemaSerializer(core_schema.dict_schema(core_schema.date_schema(), core_schema.date_schema()))
    assert v.to_python({date(2022, 12, 2): date(2022, 12, 2)}) == {date(2022, 12, 2): date(2022, 12, 2)}
    assert v.to_python({date(2022, 12, 2): date(2022, 12, 2)}, mode='json') == {'2022-12-02': '2022-12-02'}
    assert v.to_json({date(2022, 12, 2): date(2022, 12, 2)}) == b'{"2022-12-02":"2022-12-02"}'


def test_time():
    v = SchemaSerializer(core_schema.time_schema())
    assert v.to_python(time(12, 13, 14)) == time(12, 13, 14)

    assert v.to_python(time(12, 13, 14), mode='json') == '12:13:14'
    assert v.to_python(time(12, 13, 14, 123456), mode='json') == '12:13:14.123456'

    assert v.to_json(time(12, 13, 14)) == b'"12:13:14"'
    assert v.to_json(time(12, 13, 14, 123_456)) == b'"12:13:14.123456"'
    assert v.to_json(time(12, 13, 14, 123)) == b'"12:13:14.000123"'
    assert v.to_json(time(12, 13, 14, 123_000)) == b'"12:13:14.123000"'


def test_time_key():
    v = SchemaSerializer(core_schema.dict_schema(core_schema.time_schema(), core_schema.time_schema()))
    assert v.to_python({time(12, 13, 14): time(12, 13, 14)}) == {time(12, 13, 14): time(12, 13, 14)}
    assert v.to_python({time(12, 13, 14): time(12, 13, 14)}, mode='json') == {'12:13:14': '12:13:14'}
    assert v.to_json({time(12, 13, 14): time(12, 13, 14)}) == b'{"12:13:14":"12:13:14"}'


def test_any_datetime_key():
    v = SchemaSerializer(core_schema.dict_schema())
    input_value = {datetime(2022, 12, 2, 12, 13, 14): 1, date(2022, 12, 2): 2, time(12, 13, 14): 3}
    # assert v.to_python(input_value) == v
    assert v.to_python(input_value, mode='json') == {'2022-12-02T12:13:14': 1, '2022-12-02': 2, '12:13:14': 3}
    assert v.to_json(input_value) == b'{"2022-12-02T12:13:14":1,"2022-12-02":2,"12:13:14":3}'


def test_date_datetime_union():
    # See https://github.com/pydantic/pydantic/issues/7039#issuecomment-1671986746
    v = SchemaSerializer(core_schema.union_schema([core_schema.date_schema(), core_schema.datetime_schema()]))
    assert v.to_python(datetime(2022, 12, 2, 1)) == datetime(2022, 12, 2, 1)
    assert v.to_python(datetime(2022, 12, 2, 1), mode='json') == '2022-12-02T01:00:00'
    assert v.to_json(datetime(2022, 12, 2, 1)) == b'"2022-12-02T01:00:00"'


@pytest.mark.parametrize(
    'dt,expected_to_python,expected_to_json,expected_to_python_dict,expected_to_json_dict,mode',
    [
        (
            datetime(2024, 1, 1, 0, 0, 0),
            '2024-01-01T00:00:00',
            b'"2024-01-01T00:00:00"',
            {'2024-01-01T00:00:00': 'foo'},
            b'{"2024-01-01T00:00:00":"foo"}',
            'iso8601',
        ),
        (
            datetime(2024, 1, 1, 0, 0, 0),
            1704067200.0,
            b'1704067200.0',
            {'1704067200': 'foo'},
            b'{"1704067200":"foo"}',
            'seconds',
        ),
        (
            datetime(2024, 1, 1, 0, 0, 0),
            1704067200000.0,
            b'1704067200000.0',
            {'1704067200000': 'foo'},
            b'{"1704067200000":"foo"}',
            'milliseconds',
        ),
        (
            datetime(2024, 1, 1, 1, 1, 1, 23),
            1704070861.000023,
            b'1704070861.000023',
            {'1704070861.000023': 'foo'},
            b'{"1704070861.000023":"foo"}',
            'seconds',
        ),
        (
            datetime(2024, 1, 1, 1, 1, 1, 23),
            1704070861000.023,
            b'1704070861000.023',
            {'1704070861000.023': 'foo'},
            b'{"1704070861000.023":"foo"}',
            'milliseconds',
        ),
    ],
)
def test_config_datetime(
    dt: datetime, expected_to_python, expected_to_json, expected_to_python_dict, expected_to_json_dict, mode
):
    s = SchemaSerializer(core_schema.datetime_schema(), config={'ser_json_temporal': mode})
    assert s.to_python(dt) == dt
    assert s.to_python(dt, mode='json') == expected_to_python
    assert s.to_json(dt) == expected_to_json

    assert s.to_python({dt: 'foo'}) == {dt: 'foo'}
    with pytest.warns(
        UserWarning,
        match=(
            r'Expected `datetime` - serialized value may not be as expected '
            r"\[input_value=\{datetime\.datetime\([^)]*\): 'foo'\}, input_type=dict\]"
        ),
    ):
        assert s.to_python({dt: 'foo'}, mode='json') == expected_to_python_dict
    with pytest.warns(
        UserWarning,
        match=(
            r'Expected `datetime` - serialized value may not be as expected '
            r"\[input_value=\{datetime\.datetime\([^)]*\): 'foo'\}, input_type=dict\]"
        ),
    ):
        assert s.to_json({dt: 'foo'}) == expected_to_json_dict


@pytest.mark.parametrize(
    'dt,expected_to_python,expected_to_json,expected_to_python_dict,expected_to_json_dict,mode',
    [
        (
            date(2024, 1, 1),
            '2024-01-01',
            b'"2024-01-01"',
            {'2024-01-01': 'foo'},
            b'{"2024-01-01":"foo"}',
            'iso8601',
        ),
        (
            date(2024, 1, 1),
            1704067200.0,
            b'1704067200.0',
            {'1704067200': 'foo'},
            b'{"1704067200":"foo"}',
            'seconds',
        ),
        (
            date(2024, 1, 1),
            1704067200000.0,
            b'1704067200000.0',
            {'1704067200000': 'foo'},
            b'{"1704067200000":"foo"}',
            'milliseconds',
        ),
    ],
)
def test_config_date(
    dt: date, expected_to_python, expected_to_json, expected_to_python_dict, expected_to_json_dict, mode
):
    s = SchemaSerializer(core_schema.date_schema(), config={'ser_json_temporal': mode})
    assert s.to_python(dt) == dt
    assert s.to_python(dt, mode='json') == expected_to_python
    assert s.to_json(dt) == expected_to_json

    assert s.to_python({dt: 'foo'}) == {dt: 'foo'}
    with pytest.warns(
        UserWarning,
        match=(
            r'Expected `date` - serialized value may not be as expected '
            r"\[input_value=\{datetime\.date\([^)]*\): 'foo'\}, input_type=dict\]"
        ),
    ):
        assert s.to_python({dt: 'foo'}, mode='json') == expected_to_python_dict
    with pytest.warns(
        UserWarning,
        match=(
            r'Expected `date` - serialized value may not be as expected '
            r"\[input_value=\{datetime\.date\([^)]*\): 'foo'\}, input_type=dict\]"
        ),
    ):
        assert s.to_json({dt: 'foo'}) == expected_to_json_dict


@pytest.mark.parametrize(
    't,expected_to_python,expected_to_json,expected_to_python_dict,expected_to_json_dict,mode',
    [
        (
            time(3, 14, 1, 59263),
            '03:14:01.059263',
            b'"03:14:01.059263"',
            {'03:14:01.059263': 'foo'},
            b'{"03:14:01.059263":"foo"}',
            'iso8601',
        ),
        (
            time(3, 14, 1, 59263),
            11641.059263,
            b'11641.059263',
            {'11641.059263': 'foo'},
            b'{"11641.059263":"foo"}',
            'seconds',
        ),
        (
            time(3, 14, 1, 59263),
            11641059.263,
            b'11641059.263',
            {'11641059.263': 'foo'},
            b'{"11641059.263":"foo"}',
            'milliseconds',
        ),
    ],
)
def test_config_time(
    t: date, expected_to_python, expected_to_json, expected_to_python_dict, expected_to_json_dict, mode
):
    s = SchemaSerializer(core_schema.time_schema(), config={'ser_json_temporal': mode})
    assert s.to_python(t) == t
    assert s.to_python(t, mode='json') == expected_to_python
    assert s.to_json(t) == expected_to_json

    assert s.to_python({t: 'foo'}) == {t: 'foo'}
    with pytest.warns(
        UserWarning,
        match=(
            r'Expected `time` - serialized value may not be as expected '
            r"\[input_value=\{datetime\.time\([^)]*\): 'foo'\}, input_type=dict\]"
        ),
    ):
        assert s.to_python({t: 'foo'}, mode='json') == expected_to_python_dict
    with pytest.warns(
        UserWarning,
        match=(
            r'Expected `time` - serialized value may not be as expected '
            r"\[input_value=\{datetime\.time\([^)]*\): 'foo'\}, input_type=dict\]"
        ),
    ):
        assert s.to_json({t: 'foo'}) == expected_to_json_dict



import pytest

from pydantic_core import ArgsKwargs, ValidationError
from pydantic_core import core_schema as cs

from ...conftest import PyAndJson


@pytest.mark.parametrize(
    ['input_value', 'expected'],
    (
        [ArgsKwargs(()), ((), {})],
        [ArgsKwargs((), {'a': 1, 'b': 2}), ((), {'a': 1, 'b': 2})],
        [{}, ((), {})],
        [{'kwargs': {'a': 1, 'b': 2}}, ((), {'a': 1, 'b': 2})],
    ),
)
def test_var_kwargs(py_and_json: PyAndJson, input_value, expected) -> None:
    """Test valid inputs against var-args parameters (uniform):

    ```python
    def func(**kwargs: int):
        ...
    ```
    """
    v = py_and_json(
        cs.arguments_v3_schema(
            [
                cs.arguments_v3_parameter(name='kwargs', schema=cs.int_schema(), mode='var_kwargs_uniform'),
            ]
        )
    )

    assert v.validate_test(input_value) == expected


@pytest.mark.parametrize(
    ['input_value', 'err_loc'],
    (
        [ArgsKwargs((), {'a': 'not_an_int'}), ('a',)],
        [ArgsKwargs((), {'a': 1, 'b': 'not_an_int'}), ('b',)],
        [{'kwargs': {'a': 'not_an_int'}}, ('kwargs', 'a')],
        [{'kwargs': {'a': 1, 'b': 'not_an_int'}}, ('kwargs', 'b')],
    ),
)
def test_var_kwargs_validation_error(py_and_json: PyAndJson, input_value, err_loc) -> None:
    """Test invalid inputs against var-args parameters (uniform):

    ```python
    def func(**kwargs: int):
        ...

    func(a='not_an_int')
    ```
    """
    v = py_and_json(
        cs.arguments_v3_schema(
            [
                cs.arguments_v3_parameter(name='kwargs', schema=cs.int_schema(), mode='var_kwargs_uniform'),
            ]
        )
    )

    with pytest.raises(ValidationError) as exc_info:
        v.validate_test(input_value)

    error = exc_info.value.errors()[0]

    assert error['type'] == 'int_parsing'
    assert error['loc'] == err_loc


def test_var_kwargs_invalid_dict(py_and_json: PyAndJson) -> None:
    """Test invalid dict-like input against var-kwargs parameters in mapping validation mode."""
    v = py_and_json(
        cs.arguments_v3_schema(
            [
                cs.arguments_v3_parameter(name='kwargs', schema=cs.int_schema(), mode='var_kwargs_uniform'),
            ]
        )
    )

    with pytest.raises(ValidationError) as exc_info:
        v.validate_test({'kwargs': 'not_a_dict'})

    error = exc_info.value.errors()[0]

    assert error['type'] == 'dict_type'
    assert error['loc'] == ('kwargs',)



import math

import pytest

from pydantic_core import SchemaSerializer, core_schema


@pytest.mark.parametrize(
    'value,expected',
    [
        (complex(-1.23e-4, 567.89), '-0.000123+567.89j'),
        (complex(0, -1.23), '-1.23j'),
        (complex(1.5, 0), '1.5+0j'),
        (complex(1, 2), '1+2j'),
        (complex(0, 1), '1j'),
        (complex(0, 1e-500), '0j'),
        (complex(-float('inf'), 2), '-inf+2j'),
        (complex(float('inf'), 2), 'inf+2j'),
        (complex(float('nan'), 2), 'NaN+2j'),
    ],
)
def test_complex_json(value, expected):
    v = SchemaSerializer(core_schema.complex_schema())
    c = v.to_python(value)
    c_json = v.to_python(value, mode='json')
    json_str = v.to_json(value).decode()

    assert c_json == expected
    assert json_str == f'"{expected}"'

    if math.isnan(value.imag):
        assert math.isnan(c.imag)
    else:
        assert c.imag == value.imag

    if math.isnan(value.real):
        assert math.isnan(c.real)
    else:
        assert c.imag == value.imag


def test_complex_inference() -> None:
    s = SchemaSerializer(core_schema.any_schema())
    assert s.to_python(1 + 2j) == 1 + 2j
    assert s.to_json(1 + 2j) == b'"1+2j"'



from __future__ import annotations

import dataclasses
import json
import uuid
import warnings
from decimal import Decimal
from typing import Any, ClassVar, Literal, Union

import pytest

from pydantic_core import PydanticSerializationUnexpectedValue, SchemaSerializer, core_schema


class BaseModel:
    def __init__(self, **kwargs) -> None:
        for name, value in kwargs.items():
            setattr(self, name, value)


@pytest.mark.parametrize('bool_case_label', [False, True])
@pytest.mark.parametrize('int_case_label', [False, True])
@pytest.mark.parametrize('input_value,expected_value', [(True, True), (False, False), (1, 1), (123, 123), (-42, -42)])
def test_union_bool_int(input_value, expected_value, bool_case_label, int_case_label):
    bool_case = core_schema.bool_schema() if not bool_case_label else (core_schema.bool_schema(), 'my_bool_label')
    int_case = core_schema.int_schema() if not int_case_label else (core_schema.int_schema(), 'my_int_label')
    s = SchemaSerializer(core_schema.union_schema([bool_case, int_case]))

    assert s.to_python(input_value) == expected_value
    assert s.to_python(input_value, mode='json') == expected_value
    assert s.to_json(input_value) == json.dumps(expected_value).encode()


def test_union_error():
    s = SchemaSerializer(core_schema.union_schema([core_schema.bool_schema(), core_schema.int_schema()]))

    messages = [
        "Expected `bool` - serialized value may not be as expected [input_value='a string', input_type=str]",
        "Expected `int` - serialized value may not be as expected [input_value='a string', input_type=str]",
    ]

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        s.to_python('a string')
        for m in messages:
            assert m in str(w[0].message)


class ModelA:
    def __init__(self, a, b):
        self.a = a
        self.b = b


class ModelB:
    def __init__(self, c, d):
        self.c = c
        self.d = d


@pytest.fixture(scope='module')
def model_serializer() -> SchemaSerializer:
    return SchemaSerializer(
        core_schema.union_schema(
            [
                core_schema.model_schema(
                    ModelA,
                    core_schema.model_fields_schema(
                        {
                            'a': core_schema.model_field(core_schema.bytes_schema()),
                            'b': core_schema.model_field(
                                core_schema.float_schema(
                                    serialization=core_schema.format_ser_schema('0.1f', when_used='unless-none')
                                )
                            ),
                        }
                    ),
                ),
                core_schema.model_schema(
                    ModelB,
                    core_schema.model_fields_schema(
                        {
                            'c': core_schema.model_field(core_schema.bytes_schema()),
                            'd': core_schema.model_field(
                                core_schema.float_schema(
                                    serialization=core_schema.format_ser_schema('0.2f', when_used='unless-none')
                                )
                            ),
                        }
                    ),
                ),
            ],
        )
    )


class SubclassA(ModelA):
    pass


@pytest.mark.parametrize('input_value', [ModelA(b'bite', 2.3456), SubclassA(b'bite', 2.3456)])
def test_model_a(model_serializer: SchemaSerializer, input_value):
    assert model_serializer.to_python(input_value) == {'a': b'bite', 'b': '2.3'}
    assert model_serializer.to_python(input_value, mode='json') == {'a': 'bite', 'b': '2.3'}
    assert model_serializer.to_json(input_value) == b'{"a":"bite","b":"2.3"}'


class SubclassB(ModelB):
    pass


@pytest.mark.parametrize('input_value', [ModelB(b'bite', 2.3456), SubclassB(b'bite', 2.3456)])
def test_model_b(model_serializer: SchemaSerializer, input_value):
    assert model_serializer.to_python(input_value) == {'c': b'bite', 'd': '2.35'}
    assert model_serializer.to_python(input_value, mode='json') == {'c': 'bite', 'd': '2.35'}
    assert model_serializer.to_json(input_value) == b'{"c":"bite","d":"2.35"}'


def test_keys():
    s = SchemaSerializer(
        core_schema.dict_schema(
            core_schema.union_schema(
                [
                    core_schema.int_schema(),
                    core_schema.float_schema(serialization=core_schema.format_ser_schema('0.0f')),
                ]
            ),
            core_schema.int_schema(),
        )
    )
    assert s.to_python({1: 2, 2.111: 3}) == {1: 2, 2.111: 3}
    assert s.to_python({1: 2, 2.111: 3}, mode='json') == {'1': 2, '2': 3}
    assert s.to_json({1: 2, 2.111: 3}) == b'{"1":2,"2":3}'


def test_union_of_functions():
    def repr_function(value, _info):
        if value == 'unexpected':
            raise PydanticSerializationUnexpectedValue()
        return f'func: {value!r}'

    s = SchemaSerializer(
        core_schema.union_schema(
            [
                core_schema.any_schema(
                    serialization=core_schema.plain_serializer_function_ser_schema(repr_function, info_arg=True)
                ),
                core_schema.float_schema(serialization=core_schema.format_ser_schema('_^14')),
            ]
        )
    )
    assert s.to_python('foobar') == "func: 'foobar'"
    assert s.to_python('foobar', mode='json') == "func: 'foobar'"
    assert s.to_json('foobar') == b'"func: \'foobar\'"'

    assert s.to_python('unexpected') == 'unexpected'
    assert s.to_python('unexpected', mode='json') == '__unexpected__'
    assert s.to_json('unexpected') == b'"__unexpected__"'


def test_typed_dict_literal():
    s = SchemaSerializer(
        core_schema.union_schema(
            [
                core_schema.typed_dict_schema(
                    dict(
                        pet_type=core_schema.typed_dict_field(core_schema.literal_schema(['cat'])),
                        sound=core_schema.typed_dict_field(
                            core_schema.int_schema(serialization=core_schema.format_ser_schema('04d'))
                        ),
                    )
                ),
                core_schema.typed_dict_schema(
                    dict(
                        pet_type=core_schema.typed_dict_field(core_schema.literal_schema(['dog'])),
                        sound=core_schema.typed_dict_field(
                            core_schema.float_schema(serialization=core_schema.format_ser_schema('0.3f'))
                        ),
                    )
                ),
            ]
        )
    )

    assert s.to_python(dict(pet_type='cat', sound=3), mode='json') == {'pet_type': 'cat', 'sound': '0003'}
    assert s.to_python(dict(pet_type='dog', sound=3), mode='json') == {'pet_type': 'dog', 'sound': '3.000'}


def test_typed_dict_missing():
    s = SchemaSerializer(
        core_schema.union_schema(
            [
                core_schema.typed_dict_schema(dict(foo=core_schema.typed_dict_field(core_schema.int_schema()))),
                core_schema.typed_dict_schema(
                    dict(
                        foo=core_schema.typed_dict_field(
                            core_schema.int_schema(
                                serialization=core_schema.format_ser_schema('04d', when_used='always')
                            )
                        ),
                        bar=core_schema.typed_dict_field(core_schema.int_schema()),
                    )
                ),
            ]
        )
    )

    assert s.to_python(dict(foo=1)) == {'foo': 1}
    assert s.to_python(dict(foo=1), mode='json') == {'foo': 1}
    assert s.to_json(dict(foo=1)) == b'{"foo":1}'

    assert s.to_python(dict(foo=1, bar=2)) == {'foo': '0001', 'bar': 2}
    assert s.to_python(dict(foo=1, bar=2), mode='json') == {'foo': '0001', 'bar': 2}
    assert s.to_json(dict(foo=1, bar=2)) == b'{"foo":"0001","bar":2}'


def test_typed_dict_extra():
    """
    TODO, needs tests for each case
    """
    s = SchemaSerializer(
        core_schema.union_schema(
            [
                core_schema.typed_dict_schema(
                    dict(
                        foo=core_schema.typed_dict_field(core_schema.int_schema()),
                        bar=core_schema.typed_dict_field(core_schema.int_schema()),
                    )
                ),
                core_schema.typed_dict_schema(
                    dict(
                        foo=core_schema.typed_dict_field(
                            core_schema.int_schema(serialization=core_schema.format_ser_schema('04d'))
                        )
                    )
                ),
            ]
        )
    )

    assert s.to_python(dict(foo=1, bar=2)) == {'foo': 1, 'bar': 2}
    assert s.to_python(dict(foo=1, bar=2), mode='json') == {'foo': 1, 'bar': 2}
    assert s.to_json(dict(foo=1, bar=2)) == b'{"foo":1,"bar":2}'
    assert s.to_python(dict(foo=1)) == {'foo': 1}
    assert s.to_python(dict(foo=1), mode='json') == {'foo': '0001'}
    assert s.to_json(dict(foo=1)) == b'{"foo":"0001"}'


def test_typed_dict_different_fields():
    """
    TODO, needs tests for each case
    """
    s = SchemaSerializer(
        core_schema.union_schema(
            [
                core_schema.typed_dict_schema(
                    dict(
                        foo=core_schema.typed_dict_field(core_schema.int_schema()),
                        bar=core_schema.typed_dict_field(core_schema.int_schema()),
                    )
                ),
                core_schema.typed_dict_schema(
                    dict(
                        spam=core_schema.typed_dict_field(core_schema.int_schema()),
                        ham=core_schema.typed_dict_field(
                            core_schema.int_schema(serialization=core_schema.format_ser_schema('04d'))
                        ),
                    )
                ),
            ]
        )
    )

    assert s.to_python(dict(foo=1, bar=2)) == {'foo': 1, 'bar': 2}
    assert s.to_python(dict(foo=1, bar=2), mode='json') == {'foo': 1, 'bar': 2}
    assert s.to_json(dict(foo=1, bar=2)) == b'{"foo":1,"bar":2}'
    assert s.to_python(dict(spam=1, ham=2)) == {'spam': 1, 'ham': 2}
    assert s.to_python(dict(spam=1, ham=2), mode='json') == {'spam': 1, 'ham': '0002'}
    assert s.to_json(dict(spam=1, ham=2)) == b'{"spam":1,"ham":"0002"}'


def test_dataclass_union():
    @dataclasses.dataclass
    class BaseUser:
        name: str

    @dataclasses.dataclass
    class User(BaseUser):
        surname: str

    @dataclasses.dataclass
    class DBUser(User):
        password_hash: str

    @dataclasses.dataclass
    class Item:
        name: str
        price: float

    user_schema = core_schema.dataclass_schema(
        User,
        core_schema.dataclass_args_schema(
            'User',
            [
                core_schema.dataclass_field(name='name', schema=core_schema.str_schema()),
                core_schema.dataclass_field(name='surname', schema=core_schema.str_schema()),
            ],
        ),
        ['name', 'surname'],
    )
    item_schema = core_schema.dataclass_schema(
        Item,
        core_schema.dataclass_args_schema(
            'Item',
            [
                core_schema.dataclass_field(name='name', schema=core_schema.str_schema()),
                core_schema.dataclass_field(name='price', schema=core_schema.float_schema()),
            ],
        ),
        ['name', 'price'],
    )
    s = SchemaSerializer(core_schema.union_schema([user_schema, item_schema]))
    assert s.to_python(User(name='foo', surname='bar')) == {'name': 'foo', 'surname': 'bar'}
    assert s.to_python(DBUser(name='foo', surname='bar', password_hash='x')) == {'name': 'foo', 'surname': 'bar'}
    assert s.to_json(DBUser(name='foo', surname='bar', password_hash='x')) == b'{"name":"foo","surname":"bar"}'


def test_model_union():
    class BaseUser:
        def __init__(self, name: str):
            self.name = name

    class User(BaseUser):
        def __init__(self, name: str, surname: str):
            super().__init__(name)
            self.surname = surname

    class DBUser(User):
        def __init__(self, name: str, surname: str, password_hash: str):
            super().__init__(name, surname)
            self.password_hash = password_hash

    class Item:
        def __init__(self, name: str, price: float):
            self.name = name
            self.price = price

    user_schema = core_schema.model_schema(
        User,
        core_schema.model_fields_schema(
            {
                'name': core_schema.model_field(schema=core_schema.str_schema()),
                'surname': core_schema.model_field(schema=core_schema.str_schema()),
            }
        ),
    )
    item_schema = core_schema.model_schema(
        Item,
        core_schema.model_fields_schema(
            {
                'name': core_schema.model_field(schema=core_schema.str_schema()),
                'price': core_schema.model_field(schema=core_schema.float_schema()),
            }
        ),
    )
    s = SchemaSerializer(core_schema.union_schema([user_schema, item_schema]))
    assert s.to_python(User(name='foo', surname='bar')) == {'name': 'foo', 'surname': 'bar'}
    assert s.to_python(DBUser(name='foo', surname='bar', password_hash='x')) == {'name': 'foo', 'surname': 'bar'}
    assert s.to_json(DBUser(name='foo', surname='bar', password_hash='x')) == b'{"name":"foo","surname":"bar"}'


@pytest.mark.parametrize(('data', 'json_value'), [(False, 'false'), ('abc', '"abc"')])
def test_union_literal_with_other_type(data, json_value):
    class Model(BaseModel):
        value: Union[Literal[False], str]
        value_types_reversed: Union[str, Literal[False]]

    s = SchemaSerializer(
        core_schema.model_schema(
            Model,
            core_schema.model_fields_schema(
                {
                    'value': core_schema.model_field(
                        core_schema.union_schema([core_schema.literal_schema([False]), core_schema.str_schema()])
                    ),
                    'value_types_reversed': core_schema.model_field(
                        core_schema.union_schema([core_schema.str_schema(), core_schema.literal_schema([False])])
                    ),
                }
            ),
        )
    )

    m = Model(value=data, value_types_reversed=data)

    assert s.to_python(m) == {'value': data, 'value_types_reversed': data}
    assert s.to_json(m) == f'{{"value":{json_value},"value_types_reversed":{json_value}}}'.encode()


def test_union_serializes_model_subclass_from_definition() -> None:
    class BaseModel:
        __slots__ = '__dict__', '__pydantic_fields_set__', '__pydantic_extra__', '__pydantic_private__'

        def __init__(self, **kwargs: Any):
            for key, value in kwargs.items():
                setattr(self, key, value)

    class User(BaseModel):
        name: str

    class DBUser(User):
        password: str
        __pydantic_serializer__: ClassVar[SchemaSerializer]

    DBUser.__pydantic_serializer__ = SchemaSerializer(
        core_schema.model_schema(
            DBUser,
            core_schema.model_fields_schema(
                {
                    'name': core_schema.model_field(core_schema.str_schema()),
                    'password': core_schema.model_field(core_schema.str_schema()),
                }
            ),
        )
    )

    class Item(BaseModel):
        price: float

    s = SchemaSerializer(
        core_schema.definitions_schema(
            core_schema.union_schema(
                [core_schema.definition_reference_schema('User'), core_schema.definition_reference_schema('Item')]
            ),
            [
                core_schema.model_schema(
                    User,
                    core_schema.model_fields_schema({'name': core_schema.model_field(core_schema.str_schema())}),
                    ref='User',
                ),
                core_schema.model_schema(
                    Item,
                    core_schema.model_fields_schema({'price': core_schema.model_field(core_schema.float_schema())}),
                    ref='Item',
                ),
            ],
        )
    )

    assert s.to_python(DBUser(name='John', password='secret')) == {'name': 'John'}


def test_union_serializes_list_of_model_subclass_from_definition() -> None:
    class BaseModel:
        __slots__ = '__dict__', '__pydantic_fields_set__', '__pydantic_extra__', '__pydantic_private__'

        def __init__(self, **kwargs: Any):
            for key, value in kwargs.items():
                setattr(self, key, value)

    class User(BaseModel):
        name: str

    class DBUser(User):
        password: str
        __pydantic_serializer__: ClassVar[SchemaSerializer]

    DBUser.__pydantic_serializer__ = SchemaSerializer(
        core_schema.model_schema(
            DBUser,
            core_schema.model_fields_schema(
                {
                    'name': core_schema.model_field(core_schema.str_schema()),
                    'password': core_schema.model_field(core_schema.str_schema()),
                }
            ),
        )
    )

    class Item(BaseModel):
        price: float

    s = SchemaSerializer(
        core_schema.definitions_schema(
            core_schema.union_schema(
                [
                    core_schema.list_schema(core_schema.definition_reference_schema('User'), strict=False),
                    core_schema.list_schema(core_schema.definition_reference_schema('Item'), strict=False),
                ]
            ),
            [
                core_schema.model_schema(
                    User,
                    core_schema.model_fields_schema({'name': core_schema.model_field(core_schema.str_schema())}),
                    ref='User',
                ),
                core_schema.model_schema(
                    Item,
                    core_schema.model_fields_schema({'price': core_schema.model_field(core_schema.float_schema())}),
                    ref='Item',
                ),
            ],
        )
    )

    assert s.to_python([DBUser(name='John', password='secret')]) == [{'name': 'John'}]


EXAMPLE_UUID = uuid.uuid4()


class IntSubclass(int):
    pass


@pytest.mark.parametrize('reverse', [False, True])
@pytest.mark.parametrize(
    'core_schema_left,core_schema_right,input_value,expected_value',
    [
        (core_schema.int_schema(), core_schema.bool_schema(), True, True),
        (core_schema.int_schema(), core_schema.bool_schema(), 1, 1),
        (core_schema.str_schema(), core_schema.int_schema(), 1, 1),
        (core_schema.str_schema(), core_schema.int_schema(), '1', '1'),
        (core_schema.int_schema(), core_schema.bool_schema(), IntSubclass(1), 1),
        (
            core_schema.decimal_schema(),
            core_schema.int_schema(),
            Decimal('1'),
            Decimal('1'),
        ),
        (core_schema.decimal_schema(), core_schema.int_schema(), 1, 1),
        (
            core_schema.decimal_schema(),
            core_schema.float_schema(),
            Decimal('1.'),
            Decimal('1.'),
        ),
        (
            core_schema.decimal_schema(),
            core_schema.str_schema(),
            Decimal('_1'),
            Decimal('_1'),
        ),
        (
            core_schema.decimal_schema(),
            core_schema.str_schema(),
            '_1',
            '_1',
        ),
        (
            core_schema.uuid_schema(),
            core_schema.str_schema(),
            EXAMPLE_UUID,
            EXAMPLE_UUID,
        ),
        (
            core_schema.uuid_schema(),
            core_schema.str_schema(),
            str(EXAMPLE_UUID),
            str(EXAMPLE_UUID),
        ),
    ],
)
def test_union_serializer_picks_exact_type_over_subclass(
    core_schema_left, core_schema_right, input_value, expected_value, reverse
):
    s = SchemaSerializer(
        core_schema.union_schema(
            [core_schema_right, core_schema_left] if reverse else [core_schema_left, core_schema_right]
        )
    )
    assert s.to_python(input_value) == expected_value


@pytest.mark.parametrize('reverse', [False, True])
@pytest.mark.parametrize(
    'core_schema_left,core_schema_right,input_value,expected_value',
    [
        (core_schema.int_schema(), core_schema.bool_schema(), True, True),
        (core_schema.int_schema(), core_schema.bool_schema(), 1, 1),
        (core_schema.str_schema(), core_schema.int_schema(), 1, 1),
        (core_schema.str_schema(), core_schema.int_schema(), '1', '1'),
        (core_schema.int_schema(), core_schema.bool_schema(), IntSubclass(1), 1),
        (
            core_schema.decimal_schema(),
            core_schema.int_schema(),
            Decimal('1'),
            '1',
        ),
        (core_schema.decimal_schema(), core_schema.int_schema(), 1, 1),
        (
            core_schema.decimal_schema(),
            core_schema.float_schema(),
            Decimal('1.'),
            '1',
        ),
        (
            core_schema.decimal_schema(),
            core_schema.str_schema(),
            Decimal('_1'),
            '1',
        ),
        (
            core_schema.decimal_schema(),
            core_schema.str_schema(),
            '_1',
            '_1',
        ),
    ],
)
def test_union_serializer_picks_exact_type_over_subclass_json(
    core_schema_left, core_schema_right, input_value, expected_value, reverse
):
    s = SchemaSerializer(
        core_schema.union_schema(
            [core_schema_right, core_schema_left] if reverse else [core_schema_left, core_schema_right]
        )
    )
    assert s.to_python(input_value, mode='json') == expected_value
    assert s.to_json(input_value) == json.dumps(expected_value).encode()


def test_tagged_union() -> None:
    @dataclasses.dataclass
    class ModelA:
        field: int
        tag: Literal['a'] = 'a'

    @dataclasses.dataclass
    class ModelB:
        field: int
        tag: Literal['b'] = 'b'

    s = SchemaSerializer(
        core_schema.tagged_union_schema(
            choices={
                'a': core_schema.dataclass_schema(
                    ModelA,
                    core_schema.dataclass_args_schema(
                        'ModelA',
                        [
                            core_schema.dataclass_field(name='field', schema=core_schema.int_schema()),
                            core_schema.dataclass_field(name='tag', schema=core_schema.literal_schema(['a'])),
                        ],
                    ),
                    ['field', 'tag'],
                ),
                'b': core_schema.dataclass_schema(
                    ModelB,
                    core_schema.dataclass_args_schema(
                        'ModelB',
                        [
                            core_schema.dataclass_field(name='field', schema=core_schema.int_schema()),
                            core_schema.dataclass_field(name='tag', schema=core_schema.literal_schema(['b'])),
                        ],
                    ),
                    ['field', 'tag'],
                ),
            },
            discriminator='tag',
        )
    )

    assert 'TaggedUnionSerializer' in repr(s)

    model_a = ModelA(field=1)
    model_b = ModelB(field=1)
    assert s.to_python(model_a) == {'field': 1, 'tag': 'a'}
    assert s.to_python(model_b) == {'field': 1, 'tag': 'b'}


def test_union_float_int() -> None:
    s = SchemaSerializer(core_schema.union_schema([core_schema.float_schema(), core_schema.int_schema()]))

    assert s.to_python(1) == 1
    assert json.loads(s.to_json(1)) == 1

    s = SchemaSerializer(core_schema.union_schema([core_schema.int_schema(), core_schema.float_schema()]))

    assert s.to_python(1) == 1
    assert json.loads(s.to_json(1)) == 1


def test_custom_serializer() -> None:
    s = SchemaSerializer(
        core_schema.union_schema(
            [
                core_schema.dict_schema(
                    keys_schema=core_schema.any_schema(),
                    values_schema=core_schema.any_schema(),
                    serialization=core_schema.plain_serializer_function_ser_schema(lambda x: x['id']),
                ),
                core_schema.list_schema(
                    items_schema=core_schema.dict_schema(
                        keys_schema=core_schema.any_schema(),
                        values_schema=core_schema.any_schema(),
                        serialization=core_schema.plain_serializer_function_ser_schema(lambda x: x['id']),
                    )
                ),
            ]
        )
    )
    print(s)
    assert s.to_python([{'id': 1}, {'id': 2}]) == [1, 2]
    assert s.to_python({'id': 1}) == 1


def test_tagged_union_with_aliases() -> None:
    @dataclasses.dataclass
    class ModelA:
        field: int
        tag: Literal['a'] = 'a'

    @dataclasses.dataclass
    class ModelB:
        field: int
        tag: Literal['b'] = 'b'

    s = SchemaSerializer(
        core_schema.tagged_union_schema(
            choices={
                'a': core_schema.dataclass_schema(
                    ModelA,
                    core_schema.dataclass_args_schema(
                        'ModelA',
                        [
                            core_schema.dataclass_field(name='field', schema=core_schema.int_schema()),
                            core_schema.dataclass_field(
                                name='tag',
                                schema=core_schema.literal_schema(['a']),
                                validation_alias='TAG',
                                serialization_alias='TAG',
                            ),
                        ],
                    ),
                    ['field', 'tag'],
                ),
                'b': core_schema.dataclass_schema(
                    ModelB,
                    core_schema.dataclass_args_schema(
                        'ModelB',
                        [
                            core_schema.dataclass_field(name='field', schema=core_schema.int_schema()),
                            core_schema.dataclass_field(
                                name='tag',
                                schema=core_schema.literal_schema(['b']),
                                validation_alias='TAG',
                                serialization_alias='TAG',
                            ),
                        ],
                    ),
                    ['field', 'tag'],
                ),
            },
            discriminator=[['tag'], ['TAG']],
        )
    )

    assert 'TaggedUnionSerializer' in repr(s)

    model_a = ModelA(field=1)
    model_b = ModelB(field=1)
    assert s.to_python(model_a, by_alias=True) == {'field': 1, 'TAG': 'a'}
    assert s.to_python(model_b, by_alias=True) == {'field': 1, 'TAG': 'b'}


def test_union_model_wrap_serializer():
    def wrap_serializer(value, handler):
        return handler(value)

    class Data:
        pass

    class ModelA:
        a: Data

    class ModelB:
        a: Data

    model_serializer = SchemaSerializer(
        core_schema.union_schema(
            [
                core_schema.model_schema(
                    ModelA,
                    core_schema.model_fields_schema(
                        {
                            'a': core_schema.model_field(
                                core_schema.model_schema(
                                    Data,
                                    core_schema.model_fields_schema({}),
                                )
                            ),
                        },
                    ),
                    serialization=core_schema.wrap_serializer_function_ser_schema(wrap_serializer),
                ),
                core_schema.model_schema(
                    ModelB,
                    core_schema.model_fields_schema(
                        {
                            'a': core_schema.model_field(
                                core_schema.model_schema(
                                    Data,
                                    core_schema.model_fields_schema({}),
                                )
                            ),
                        },
                    ),
                    serialization=core_schema.wrap_serializer_function_ser_schema(wrap_serializer),
                ),
            ],
        )
    )

    input_value = ModelA()
    input_value.a = Data()

    assert model_serializer.to_python(input_value) == {'a': {}}
    assert model_serializer.to_python(input_value, mode='json') == {'a': {}}
    assert model_serializer.to_json(input_value) == b'{"a":{}}'

    # add some additional attribute, should be ignored & not break serialization

    input_value.a._a = 'foo'

    assert model_serializer.to_python(input_value) == {'a': {}}
    assert model_serializer.to_python(input_value, mode='json') == {'a': {}}
    assert model_serializer.to_json(input_value) == b'{"a":{}}'


class ModelDog:
    def __init__(self, type_: Literal['dog']) -> None:
        self.type_ = 'dog'


class ModelCat:
    def __init__(self, type_: Literal['cat']) -> None:
        self.type_ = 'cat'


class ModelAlien:
    def __init__(self, type_: Literal['alien']) -> None:
        self.type_ = 'alien'


@pytest.fixture
def model_a_b_union_schema() -> core_schema.UnionSchema:
    return core_schema.union_schema(
        [
            core_schema.model_schema(
                cls=ModelA,
                schema=core_schema.model_fields_schema(
                    fields={
                        'a': core_schema.model_field(core_schema.str_schema()),
                        'b': core_schema.model_field(core_schema.str_schema()),
                    },
                ),
            ),
            core_schema.model_schema(
                cls=ModelB,
                schema=core_schema.model_fields_schema(
                    fields={
                        'c': core_schema.model_field(core_schema.str_schema()),
                        'd': core_schema.model_field(core_schema.str_schema()),
                    },
                ),
            ),
        ]
    )


@pytest.fixture
def union_of_unions_schema(model_a_b_union_schema: core_schema.UnionSchema) -> core_schema.UnionSchema:
    return core_schema.union_schema(
        [
            model_a_b_union_schema,
            core_schema.union_schema(
                [
                    core_schema.model_schema(
                        cls=ModelCat,
                        schema=core_schema.model_fields_schema(
                            fields={
                                'type_': core_schema.model_field(core_schema.literal_schema(['cat'])),
                            },
                        ),
                    ),
                    core_schema.model_schema(
                        cls=ModelDog,
                        schema=core_schema.model_fields_schema(
                            fields={
                                'type_': core_schema.model_field(core_schema.literal_schema(['dog'])),
                            },
                        ),
                    ),
                ]
            ),
        ]
    )


@pytest.mark.parametrize(
    'input,expected',
    [
        (ModelA(a='a', b='b'), {'a': 'a', 'b': 'b'}),
        (ModelB(c='c', d='d'), {'c': 'c', 'd': 'd'}),
        (ModelCat(type_='cat'), {'type_': 'cat'}),
        (ModelDog(type_='dog'), {'type_': 'dog'}),
    ],
)
def test_union_of_unions_of_models(union_of_unions_schema: core_schema.UnionSchema, input: Any, expected: Any) -> None:
    s = SchemaSerializer(union_of_unions_schema)
    assert s.to_python(input, warnings='error') == expected


def test_union_of_unions_of_models_invalid_variant(union_of_unions_schema: core_schema.UnionSchema) -> None:
    s = SchemaSerializer(union_of_unions_schema)
    # All warnings should be available
    messages = [
        'Expected `ModelA`',
        'Expected `ModelB`',
        'Expected `ModelCat`',
        'Expected `ModelDog`',
    ]

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        s.to_python(ModelAlien(type_='alien'))
        for m in messages:
            assert m in str(w[0].message)
            assert 'input_type=ModelAlien' in str(w[0].message)


@pytest.fixture
def tagged_union_of_unions_schema(model_a_b_union_schema: core_schema.UnionSchema) -> core_schema.UnionSchema:
    return core_schema.union_schema(
        [
            model_a_b_union_schema,
            core_schema.tagged_union_schema(
                discriminator='type_',
                choices={
                    'cat': core_schema.model_schema(
                        cls=ModelCat,
                        schema=core_schema.model_fields_schema(
                            fields={
                                'type_': core_schema.model_field(core_schema.literal_schema(['cat'])),
                            },
                        ),
                    ),
                    'dog': core_schema.model_schema(
                        cls=ModelDog,
                        schema=core_schema.model_fields_schema(
                            fields={
                                'type_': core_schema.model_field(core_schema.literal_schema(['dog'])),
                            },
                        ),
                    ),
                },
            ),
        ]
    )


@pytest.mark.parametrize(
    'input,expected',
    [
        (ModelA(a='a', b='b'), {'a': 'a', 'b': 'b'}),
        (ModelB(c='c', d='d'), {'c': 'c', 'd': 'd'}),
        (ModelCat(type_='cat'), {'type_': 'cat'}),
        (ModelDog(type_='dog'), {'type_': 'dog'}),
    ],
)
def test_union_of_unions_of_models_with_tagged_union(
    tagged_union_of_unions_schema: core_schema.UnionSchema, input: Any, expected: Any
) -> None:
    s = SchemaSerializer(tagged_union_of_unions_schema)
    assert s.to_python(input, warnings='error') == expected


def test_union_of_unions_of_models_with_tagged_union_invalid_variant(
    tagged_union_of_unions_schema: core_schema.UnionSchema,
) -> None:
    s = SchemaSerializer(tagged_union_of_unions_schema)
    # All warnings should be available
    messages = [
        'Expected `ModelA`',
        'Expected `ModelB`',
        'Expected `ModelCat`',
        'Expected `ModelDog`',
    ]

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        s.to_python(ModelAlien(type_='alien'))
        for m in messages:
            assert m in str(w[0].message)
            assert 'input_type=ModelAlien' in str(w[0].message)


def test_mixed_union_models_and_other_types() -> None:
    s = SchemaSerializer(
        core_schema.union_schema(
            [
                core_schema.tagged_union_schema(
                    discriminator='type_',
                    choices={
                        'cat': core_schema.model_schema(
                            cls=ModelCat,
                            schema=core_schema.model_fields_schema(
                                fields={
                                    'type_': core_schema.model_field(core_schema.literal_schema(['cat'])),
                                },
                            ),
                        ),
                        'dog': core_schema.model_schema(
                            cls=ModelDog,
                            schema=core_schema.model_fields_schema(
                                fields={
                                    'type_': core_schema.model_field(core_schema.literal_schema(['dog'])),
                                },
                            ),
                        ),
                    },
                ),
                core_schema.str_schema(),
            ]
        )
    )

    assert s.to_python(ModelCat(type_='cat'), warnings='error') == {'type_': 'cat'}
    assert s.to_python(ModelDog(type_='dog'), warnings='error') == {'type_': 'dog'}
    # note, this fails as ModelCat and ModelDog (discriminator warnings, etc), but the warnings
    # don't bubble up to this level :)
    assert s.to_python('a string', warnings='error') == 'a string'


@pytest.mark.parametrize(
    'input,expected',
    [
        ({True: '1'}, b'{"true":"1"}'),
        ({1: '1'}, b'{"1":"1"}'),
        ({2.3: '1'}, b'{"2.3":"1"}'),
        ({'a': 'b'}, b'{"a":"b"}'),
    ],
)
def test_union_of_unions_of_models_with_tagged_union_json_key_serialization(
    input: dict[bool | int | float | str, str], expected: bytes
) -> None:
    s = SchemaSerializer(
        core_schema.dict_schema(
            keys_schema=core_schema.union_schema(
                [
                    core_schema.union_schema([core_schema.bool_schema(), core_schema.int_schema()]),
                    core_schema.union_schema([core_schema.float_schema(), core_schema.str_schema()]),
                ]
            ),
            values_schema=core_schema.str_schema(),
        )
    )

    assert s.to_json(input, warnings='error') == expected


@pytest.mark.parametrize(
    'input,expected',
    [
        ({'key': True}, b'{"key":true}'),
        ({'key': 1}, b'{"key":1}'),
        ({'key': 2.3}, b'{"key":2.3}'),
        ({'key': 'a'}, b'{"key":"a"}'),
    ],
)
def test_union_of_unions_of_models_with_tagged_union_json_serialization(
    input: dict[str, bool | int | float | str], expected: bytes
) -> None:
    s = SchemaSerializer(
        core_schema.dict_schema(
            keys_schema=core_schema.str_schema(),
            values_schema=core_schema.union_schema(
                [
                    core_schema.union_schema([core_schema.bool_schema(), core_schema.int_schema()]),
                    core_schema.union_schema([core_schema.float_schema(), core_schema.str_schema()]),
                ]
            ),
        )
    )

    assert s.to_json(input, warnings='error') == expected


def test_discriminated_union_ser_with_typed_dict() -> None:
    v = SchemaSerializer(
        core_schema.tagged_union_schema(
            {
                'a': core_schema.typed_dict_schema(
                    {
                        'type': core_schema.typed_dict_field(core_schema.literal_schema(['a'])),
                        'a': core_schema.typed_dict_field(core_schema.int_schema()),
                    }
                ),
                'b': core_schema.typed_dict_schema(
                    {
                        'type': core_schema.typed_dict_field(core_schema.literal_schema(['b'])),
                        'b': core_schema.typed_dict_field(core_schema.str_schema()),
                    }
                ),
            },
            discriminator='type',
        )
    )

    assert v.to_python({'type': 'a', 'a': 1}, warnings='error') == {'type': 'a', 'a': 1}
    assert v.to_python({'type': 'b', 'b': 'foo'}, warnings='error') == {'type': 'b', 'b': 'foo'}



from __future__ import annotations as _annotations

import functools
import gc
import importlib.util
import json
import os
import re
import sys
from dataclasses import dataclass
from pathlib import Path
from time import sleep, time
from typing import Any, Callable, Literal

import hypothesis
import pytest

from pydantic_core import ArgsKwargs, CoreSchema, SchemaValidator, ValidationError
from pydantic_core.core_schema import CoreConfig, ExtraBehavior

__all__ = 'Err', 'PyAndJson', 'assert_gc', 'is_free_threaded', 'plain_repr', 'infinite_generator'

hypothesis.settings.register_profile('fast', max_examples=2)
hypothesis.settings.register_profile('slow', max_examples=1_000)
hypothesis.settings.load_profile(os.getenv('HYPOTHESIS_PROFILE', 'fast'))

try:
    is_free_threaded = not sys._is_gil_enabled()
except AttributeError:
    is_free_threaded = False


def plain_repr(obj):
    r = repr(obj)
    r = re.sub(r',\s*([)}])', r'\1', r)
    r = re.sub(r'\s+', '', r)
    return r


@dataclass
class Err:
    message: str
    errors: Any | None = None

    def __repr__(self):
        if self.errors:
            return f'Err({self.message!r}, errors={self.errors!r})'
        else:
            return f'Err({self.message!r})'


def json_default(obj):
    if isinstance(obj, ArgsKwargs):
        raise pytest.skip('JSON skipping ArgsKwargs')
    else:
        raise TypeError(f'Object of type {type(obj).__name__} is not JSON serializable')


class PyAndJsonValidator:
    def __init__(
        self,
        schema: CoreSchema,
        config: CoreConfig | None = None,
        *,
        validator_type: Literal['json', 'python'] | None = None,
    ):
        self.validator = SchemaValidator(schema, config)
        self.validator_type = validator_type

    def validate_python(self, py_input, strict: bool | None = None, context: Any = None):
        return self.validator.validate_python(py_input, strict=strict, context=context)

    def validate_json(self, json_str: str, strict: bool | None = None, context: Any = None):
        return self.validator.validate_json(json_str, strict=strict, context=context)

    def validate_test(
        self, py_input, strict: bool | None = None, context: Any = None, extra: ExtraBehavior | None = None
    ):
        if self.validator_type == 'json':
            return self.validator.validate_json(
                json.dumps(py_input, default=json_default),
                strict=strict,
                extra=extra,
                context=context,
            )
        else:
            assert self.validator_type == 'python', self.validator_type
            return self.validator.validate_python(py_input, strict=strict, context=context, extra=extra)

    def isinstance_test(self, py_input, strict: bool | None = None, context: Any = None):
        if self.validator_type == 'json':
            try:
                self.validator.validate_json(json.dumps(py_input), strict=strict, context=context)
                return True
            except ValidationError:
                return False
        else:
            assert self.validator_type == 'python', self.validator_type
            return self.validator.isinstance_python(py_input, strict=strict, context=context)


PyAndJson = type[PyAndJsonValidator]


@pytest.fixture(params=['python', 'json'])
def py_and_json(request) -> PyAndJson:
    class ChosenPyAndJsonValidator(PyAndJsonValidator):
        __init__ = functools.partialmethod(PyAndJsonValidator.__init__, validator_type=request.param)

    return ChosenPyAndJsonValidator


class StrictModeType:
    def __init__(self, schema: bool, extra: bool):
        assert schema or extra
        self.schema = schema
        self.validator_args = {'strict': True} if extra else {}


@pytest.fixture(
    params=[
        StrictModeType(schema=True, extra=False),
        StrictModeType(schema=False, extra=True),
        StrictModeType(schema=True, extra=True),
    ],
    ids=['strict-schema', 'strict-extra', 'strict-both'],
)
def strict_mode_type(request) -> StrictModeType:
    return request.param


@pytest.fixture
def tmp_work_path(tmp_path: Path):
    """
    Create a temporary working directory.
    """
    previous_cwd = Path.cwd()
    os.chdir(tmp_path)

    yield tmp_path

    os.chdir(previous_cwd)


@pytest.fixture
def import_execute(request, tmp_work_path: Path):
    def _import_execute(source: str, *, custom_module_name: str | None = None):
        module_name = custom_module_name or request.node.name

        module_path = tmp_work_path / f'{module_name}.py'
        module_path.write_text(source)
        spec = importlib.util.spec_from_file_location('__main__', str(module_path))
        module = importlib.util.module_from_spec(spec)
        try:
            spec.loader.exec_module(module)
        except KeyboardInterrupt:
            print('KeyboardInterrupt')
        else:
            return module

    return _import_execute


@pytest.fixture
def pydantic_version():
    try:
        import pydantic

        # include major and minor version only
        return '.'.join(pydantic.__version__.split('.')[:2])
    except ImportError:
        return 'latest'


def infinite_generator():
    i = 0
    while True:
        yield i
        i += 1


def assert_gc(test: Callable[[], bool], timeout: float = 10) -> None:
    """Helper to retry garbage collection until the test passes or timeout is
    reached.

    This is useful on free-threading where the GC collect call finishes before
    all cleanup is done.
    """
    start = now = time()
    while now - start < timeout:
        if test():
            return
        gc.collect()
        sleep(0.1)
        now = time()
    raise AssertionError('Timeout waiting for GC')



import pytest

from pydantic_core import ArgsKwargs, ValidationError
from pydantic_core import core_schema as cs

from ...conftest import PyAndJson


@pytest.mark.parametrize(
    ['input_value', 'expected'],
    (
        [ArgsKwargs((1, True)), ((1, True), {})],
        [ArgsKwargs((1,)), ((1,), {'b': True})],
        [ArgsKwargs((1,), {'b': True}), ((1,), {'b': True})],
        [ArgsKwargs((), {'a': 1, 'b': True}), ((), {'a': 1, 'b': True})],
        [{'a': 1, 'b': True}, ((1, True), {})],
        [{'a': 1}, ((1, True), {})],
    ),
)
def test_positional_or_keyword(py_and_json: PyAndJson, input_value, expected) -> None:
    """Test valid inputs against positional-or-keyword parameters:

    ```python
    def func(a: int, b: bool = True):
        ...
    ```
    """
    v = py_and_json(
        cs.arguments_v3_schema(
            [
                cs.arguments_v3_parameter(name='a', schema=cs.int_schema(), mode='positional_or_keyword'),
                cs.arguments_v3_parameter(
                    name='b',
                    schema=cs.with_default_schema(cs.bool_schema(), default=True),
                    mode='positional_or_keyword',
                ),
            ]
        )
    )

    assert v.validate_test(input_value) == expected


@pytest.mark.parametrize(
    ['input_value', 'err_loc'],
    (
        [ArgsKwargs(('not_an_int',), {}), (0,)],
        [ArgsKwargs((), {'a': 'not_an_int'}), ('a',)],
        [{'a': 'not_an_int'}, ('a',)],
    ),
)
def test_positional_or_keyword_validation_error(py_and_json: PyAndJson, input_value, err_loc) -> None:
    """Test invalid inputs against positional-or-keyword parameters:

    ```python
    def func(a: int):
        ...

    func('not_an_int')
    ```
    """
    v = py_and_json(
        cs.arguments_v3_schema(
            [
                cs.arguments_v3_parameter(name='a', schema=cs.int_schema(), mode='positional_or_keyword'),
            ]
        )
    )

    with pytest.raises(ValidationError) as exc_info:
        v.validate_test(input_value)

    error = exc_info.value.errors()[0]

    assert error['type'] == 'int_parsing'
    assert error['loc'] == err_loc


@pytest.mark.parametrize(
    'input_value',
    [ArgsKwargs((), {}), {}],
)
def test_positional_only_error_required(py_and_json: PyAndJson, input_value) -> None:
    """Test missing inputs against positional-or-keyword parameters:

    ```python
    def func(a: int):
        ...

    func()
    ```
    """
    v = py_and_json(
        cs.arguments_v3_schema(
            [
                cs.arguments_v3_parameter(name='a', schema=cs.int_schema(), mode='positional_or_keyword'),
            ]
        )
    )

    with pytest.raises(ValidationError) as exc_info:
        v.validate_test(input_value)

    error = exc_info.value.errors()[0]

    assert error['type'] == 'missing_argument'
    assert error['loc'] == ('a',)



import pytest

from pydantic_core import PydanticOmit, SchemaError, SchemaValidator, ValidationError, core_schema
from pydantic_core import core_schema as cs

from .conftest import PyAndJson


def test_isinstance():
    v = SchemaValidator(cs.int_schema())
    assert v.validate_python(123) == 123
    assert v.isinstance_python(123) is True
    assert v.validate_python('123') == 123
    assert v.isinstance_python('123') is True

    with pytest.raises(ValidationError, match='Input should be a valid integer'):
        v.validate_python('foo')

    assert v.isinstance_python('foo') is False


def test_isinstance_strict():
    v = SchemaValidator(cs.int_schema(strict=True))
    assert v.validate_python(123) == 123
    assert v.isinstance_python(123) is True

    with pytest.raises(ValidationError, match='Input should be a valid integer'):
        v.validate_python('123')

    assert v.isinstance_python('123') is False


def test_isinstance_forbid_extra_fn_override():
    v = SchemaValidator(cs.typed_dict_schema({'f': cs.typed_dict_field(cs.str_schema())}))

    with pytest.raises(ValidationError, match='Extra inputs are not permitted'):
        v.validate_python({'f': 'x', 'extra_field': '123'}, extra='forbid')

    assert v.isinstance_python({'f': 'x', 'extra_field': '123'}, extra='forbid') is False


def test_internal_error():
    v = SchemaValidator(
        cs.model_schema(cls=int, schema=cs.model_fields_schema(fields={'f': cs.model_field(schema=cs.int_schema())}))
    )
    with pytest.raises(AttributeError, match="'int' object has no attribute '__dict__'"):
        v.validate_python({'f': 123})

    with pytest.raises(AttributeError, match="'int' object has no attribute '__dict__'"):
        v.validate_json('{"f": 123}')

    with pytest.raises(AttributeError, match="'int' object has no attribute '__dict__'"):
        v.isinstance_python({'f': 123})


def test_omit(py_and_json: PyAndJson):
    def omit(v, info):
        if v == 'omit':
            raise PydanticOmit
        elif v == 'error':
            raise ValueError('error')
        else:
            return v

    v = py_and_json(core_schema.with_info_plain_validator_function(omit))
    assert v.validate_test('foo') == 'foo'
    if v.validator_type == 'python':
        assert v.isinstance_test('foo') is True

    if v.validator_type == 'python':
        assert v.isinstance_test('error') is False
    with pytest.raises(SchemaError, match='Uncaught Omit error, please check your usage of `default` validators.'):
        v.validate_test('omit')



import re
from enum import Enum

import pytest

from pydantic_core import SchemaValidator, ValidationError, core_schema

from ..conftest import Err, PyAndJson, plain_repr


@pytest.mark.parametrize(
    'input_value,expected',
    [
        ('{"a": 1}', {'a': 1}),
        ('"a"', 'a'),
        ('1', 1),
        ('[1, 2, 3, "4"]', [1, 2, 3, '4']),
        (
            '{1: 2}',
            Err(
                'Invalid JSON: key must be a string at line 1 column 2 [type=json_invalid,',
                [
                    {
                        'type': 'json_invalid',
                        'loc': (),
                        'msg': 'Invalid JSON: key must be a string at line 1 column 2',
                        'input': '{1: 2}',
                        'ctx': {'error': 'key must be a string at line 1 column 2'},
                    }
                ],
            ),
        ),
        (44, Err('JSON input should be string, bytes or bytearray [type=json_type, input_value=44, input_type=int')),
    ],
)
def test_any(py_and_json: PyAndJson, input_value, expected):
    v = py_and_json(core_schema.json_schema())
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)) as exc_info:
            v.validate_test(input_value)

        if expected.errors is not None:
            # debug(exc_info.value.errors(include_url=False))
            assert exc_info.value.errors(include_url=False) == expected.errors
    else:
        assert v.validate_test(input_value) == expected


@pytest.mark.parametrize(
    'input_value,expected',
    [
        pytest.param('{"a": 1}', {'a': 1}, id='str'),
        pytest.param(b'{"a": 1}', {'a': 1}, id='bytes'),
        pytest.param(
            '🐈 Hello \ud800World',
            Err(
                'Input should be a valid string, unable to parse raw data as a unicode string '
                "[type=string_unicode, input_value='🐈 Hello \\ud800World', input_type=str]"
            ),
            id='str_unicode',
        ),
        pytest.param(bytearray(b'{"a": 1}'), {'a': 1}, id='bytearray'),
        pytest.param(
            'xx',
            Err(
                "Invalid JSON: expected value at line 1 column 1 [type=json_invalid, input_value='xx', input_type=str]"
            ),
            id='str_invalid',
        ),
        pytest.param(
            b'xx',
            Err(
                'Invalid JSON: expected value at line 1 column 1 '
                "[type=json_invalid, input_value=b'xx', input_type=bytes]"
            ),
            id='bytes_invalid',
        ),
        pytest.param(
            bytearray(b'xx'),
            Err(
                'Invalid JSON: expected value at line 1 column 1 '
                "[type=json_invalid, input_value=bytearray(b'xx'), input_type=bytearray]"
            ),
            id='bytearray_invalid',
        ),
    ],
)
def test_any_python(input_value, expected):
    v = SchemaValidator(core_schema.json_schema())
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            v.validate_python(input_value)
    else:
        assert v.validate_python(input_value) == expected


@pytest.mark.parametrize(
    'input_value,expected',
    [
        ('[1]', [1]),
        ('[1, 2, 3, "4"]', [1, 2, 3, 4]),
        ('44', Err(r'Input should be a valid (list|array) \[type=list_type, input_value=44, input_type=int')),
        ('"x"', Err(r"Input should be a valid (list|array) \[type=list_type, input_value='x', input_type=str")),
        (
            '[1, 2, 3, "err"]',
            Err(
                r'Input should be a valid integer, unable to parse string as an integer \[type=int_parsing,',
                [
                    {
                        'type': 'int_parsing',
                        'loc': (3,),
                        'msg': 'Input should be a valid integer, unable to parse string as an integer',
                        'input': 'err',
                    }
                ],
            ),
        ),
    ],
)
def test_list_int(py_and_json: PyAndJson, input_value, expected):
    v = py_and_json(core_schema.json_schema(core_schema.list_schema(core_schema.int_schema())))
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=expected.message) as exc_info:
            v.validate_test(input_value)

        if expected.errors is not None:
            # debug(exc_info.value.errors(include_url=False))
            assert exc_info.value.errors(include_url=False) == expected.errors
    else:
        assert v.validate_test(input_value) == expected


def test_dict_key(py_and_json: PyAndJson):
    v = py_and_json(
        core_schema.dict_schema(
            core_schema.json_schema(core_schema.tuple_positional_schema([core_schema.int_schema()])),
            core_schema.int_schema(),
        )
    )
    assert v.validate_test({'[1]': 4}) == {(1,): 4}
    with pytest.raises(ValidationError) as exc_info:
        v.validate_test({'x': 4})
    # insert_assert(exc_info.value.errors(include_url=False))
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'json_invalid',
            'loc': ('x', '[key]'),
            'msg': 'Invalid JSON: expected value at line 1 column 1',
            'input': 'x',
            'ctx': {'error': 'expected value at line 1 column 1'},
        }
    ]


def test_enum() -> None:
    class MyEnum(Enum):
        a = 'a'
        b = 'b'

    enum_schema = core_schema.lax_or_strict_schema(
        core_schema.no_info_after_validator_function(MyEnum, core_schema.str_schema()),
        core_schema.is_instance_schema(MyEnum),
    )
    v = core_schema.json_schema(enum_schema)
    v = SchemaValidator(v)
    assert v.validate_python('"a"') == MyEnum.a
    assert v.validate_python('"b"') == MyEnum.b
    with pytest.raises(ValidationError):
        v.validate_python('"c"')


def test_any_schema_no_schema():
    v = SchemaValidator(core_schema.json_schema())
    assert 'validator:None' in plain_repr(v)
    v = SchemaValidator(core_schema.json_schema(core_schema.any_schema()))
    assert 'validator:None' in plain_repr(v)
    v = SchemaValidator(core_schema.json_schema(core_schema.int_schema()))
    assert 'validator:Some(' in plain_repr(v)



import json
from enum import Enum

import pytest

from pydantic_core import PydanticSerializationError, SchemaSerializer, core_schema


def test_str():
    v = SchemaSerializer(core_schema.str_schema())
    assert v.to_python('foobar') == 'foobar'
    assert v.to_python('emoji 💩') == 'emoji 💩'
    assert v.to_json('foobar') == b'"foobar"'
    assert v.to_json('foobar', indent=2) == b'"foobar"'
    assert v.to_json('emoji 💩') == b'"emoji \xf0\x9f\x92\xa9"'
    assert json.loads(v.to_json('emoji 💩')) == 'emoji 💩'

    assert v.to_python('foobar', mode='json') == 'foobar'

    json_emoji = v.to_json('emoji 💩')
    # note! serde_json serializes unicode characters differently to json.dumps, but it's still valid JSON
    assert json_emoji == b'"emoji \xf0\x9f\x92\xa9"'
    assert json.loads(json_emoji) == 'emoji 💩'


# Tests borrowed from:
# - https://github.com/python/cpython/blob/d87e7f35/Lib/test/test_json/test_encode_basestring_ascii.py
# - https://github.com/python/cpython/blob/d87e7f35/Lib/test/test_json/test_unicode.py
@pytest.mark.parametrize(
    ['input', 'expected'],
    [
        (
            '/\\"\ucafe\ubabe\uab98\ufcde\ubcda\uef4a\x08\x0c\n\r\t`1~!@#$%^&*()_+-=[]{}|;:\',./<>?',
            '"/\\\\\\"\\ucafe\\ubabe\\uab98\\ufcde\\ubcda\\uef4a\\b\\f\\n\\r\\t`1~!@#$%^&*()_+-=[]{}|;:\',./<>?"',
        ),
        ('\u0123\u4567\u89ab\ucdef\uabcd\uef4a', '"\\u0123\\u4567\\u89ab\\ucdef\\uabcd\\uef4a"'),
        ('controls', '"controls"'),
        ('\x08\x0c\n\r\t', '"\\b\\f\\n\\r\\t"'),
        (
            '{"object with 1 member":["array with 1 element"]}',
            '"{\\"object with 1 member\\":[\\"array with 1 element\\"]}"',
        ),
        (' s p a c e d ', '" s p a c e d "'),
        ('\U0001d120', '"\\ud834\\udd20"'),
        ('\u03b1\u03a9', '"\\u03b1\\u03a9"'),
        ("`1~!@#$%^&*()_+-={':[,]}|;.</>?", '"`1~!@#$%^&*()_+-={\':[,]}|;.</>?"'),
        ('\x08\x0c\n\r\t', '"\\b\\f\\n\\r\\t"'),
        ('\u0123\u4567\u89ab\ucdef\uabcd\uef4a', '"\\u0123\\u4567\\u89ab\\ucdef\\uabcd\\uef4a"'),
        ('\N{GREEK SMALL LETTER ALPHA}\N{GREEK CAPITAL LETTER OMEGA}', '"\\u03b1\\u03a9"'),
        ('\U0001d120', '"\\ud834\\udd20"'),
    ],
)
def test_str_ensure_ascii(input: str, expected: str) -> None:
    v = SchemaSerializer(core_schema.str_schema())
    assert v.to_json(input, ensure_ascii=True).decode('utf-8') == expected


def test_huge_str():
    v = SchemaSerializer(core_schema.int_schema())
    msg = r"Expected `int` - serialized value may not be as expected \[input_value='123456789012345678901234...89012345678901234567890', input_type=str\]"
    with pytest.warns(UserWarning, match=msg):
        v.to_python(
            '12345678901234567890123456789012345678901234567890123456789012345678901234567890\
                           12345678901234567890123456789012345678901234567890123456789012345678901234567890\
                           12345678901234567890123456789012345678901234567890123456789012345678901234567890\
                           12345678901234567890123456789012345678901234567890123456789012345678901234567890'
        )


def test_str_fallback():
    s = SchemaSerializer(core_schema.str_schema())
    assert s.to_python(None) is None
    assert s.to_python(None, mode='json') is None
    assert s.to_json(None) == b'null'
    with pytest.warns(
        UserWarning,
        match=r'Expected `str` - serialized value may not be as expected \[input_value=123, input_type=int\]',
    ):
        assert s.to_python(123) == 123
    with pytest.warns(
        UserWarning,
        match=r'Expected `str` - serialized value may not be as expected \[input_value=123, input_type=int\]',
    ):
        assert s.to_python(123, mode='json') == 123
    with pytest.warns(
        UserWarning,
        match=r'Expected `str` - serialized value may not be as expected \[input_value=123, input_type=int\]',
    ):
        assert s.to_json(123) == b'123'
    with pytest.warns(
        UserWarning,
        match=r'Expected `str` - serialized value may not be as expected \[input_value=123, input_type=int\]',
    ):
        assert s.to_python(123, warnings='warn') == 123
    with pytest.warns(
        UserWarning,
        match=r'Expected `str` - serialized value may not be as expected \[input_value=123, input_type=int\]',
    ):
        assert s.to_python(123, mode='json', warnings='warn') == 123
    with pytest.warns(
        UserWarning,
        match=r'Expected `str` - serialized value may not be as expected \[input_value=123, input_type=int\]',
    ):
        assert s.to_json(123, warnings='warn') == b'123'
    with pytest.warns(
        UserWarning,
        match=r'Expected `str` - serialized value may not be as expected \[input_value=123, input_type=int\]',
    ):
        assert s.to_python(123, warnings=True) == 123
    with pytest.warns(
        UserWarning,
        match=r'Expected `str` - serialized value may not be as expected \[input_value=123, input_type=int\]',
    ):
        assert s.to_python(123, mode='json', warnings=True) == 123
    with pytest.warns(
        UserWarning,
        match=r'Expected `str` - serialized value may not be as expected \[input_value=123, input_type=int\]',
    ):
        assert s.to_json(123, warnings=True) == b'123'


def test_str_no_warnings():
    s = SchemaSerializer(core_schema.str_schema())
    assert s.to_python(123, warnings=False) == 123
    assert s.to_python(123, warnings='none') == 123
    assert s.to_python(123, mode='json', warnings=False) == 123
    assert s.to_python(123, mode='json', warnings='none') == 123
    assert s.to_json(123, warnings=False) == b'123'
    assert s.to_json(123, warnings='none') == b'123'


def test_str_errors():
    s = SchemaSerializer(core_schema.str_schema())
    with pytest.raises(
        PydanticSerializationError,
        match=r'Expected `str` - serialized value may not be as expected \[input_value=123, input_type=int\]',
    ):
        assert s.to_python(123, warnings='error') == 123
    with pytest.raises(
        PydanticSerializationError,
        match=r'Expected `str` - serialized value may not be as expected \[input_value=123, input_type=int\]',
    ):
        assert s.to_python(123, mode='json', warnings='error') == 123
    with pytest.raises(
        PydanticSerializationError,
        match=r'Expected `str` - serialized value may not be as expected \[input_value=123, input_type=int\]',
    ):
        assert s.to_json(123, warnings='error') == b'123'


class StrSubclass(str):
    pass


class BasicClass:
    pass


class StrMixin(str, BasicClass):
    pass


class StrEnum(str, Enum):
    foo = 'foo-value'
    bar = 'bar-value'


@pytest.mark.parametrize('schema_type', ['str', 'any'])
@pytest.mark.parametrize(
    'input_value,expected', [(StrSubclass('foo'), 'foo'), (StrMixin('foo'), 'foo'), (StrEnum.foo, 'foo-value')]
)
def test_subclass_str(schema_type, input_value, expected):
    s = SchemaSerializer({'type': schema_type})
    v = s.to_python(input_value)
    assert v == input_value
    assert type(v) == type(input_value)

    v = s.to_python(input_value, mode='json')
    assert v == expected
    assert type(v) == str

    assert s.to_json(input_value) == json.dumps(expected).encode('utf-8')



import math
import re

import pytest
from dirty_equals import FunctionCheck, HasAttributes, IsInstance

from pydantic_core import CoreConfig, SchemaValidator, ValidationError
from pydantic_core import core_schema as cs

from .conftest import Err, plain_repr


def test_on_field():
    v = SchemaValidator(cs.str_schema(min_length=2, max_length=5))
    r = plain_repr(v)
    assert 'min_length:Some(2)' in r
    assert 'max_length:Some(5)' in r
    assert v.isinstance_python('test') is True
    assert v.isinstance_python('test long') is False


def test_on_config():
    v = SchemaValidator(cs.str_schema(), config=CoreConfig(str_max_length=5))
    assert 'max_length:Some(5)' in plain_repr(v)
    assert v.isinstance_python('test') is True
    assert v.isinstance_python('test long') is False


def test_field_priority_arg():
    v = SchemaValidator(cs.str_schema(max_length=5), config=CoreConfig(str_max_length=10))
    assert 'max_length:Some(5)' in plain_repr(v)
    assert v.isinstance_python('test') is True
    assert v.isinstance_python('test long') is False


class MyModel:
    # this is not required, but it avoids `__pydantic_fields_set__` being included in `__dict__`
    __slots__ = '__dict__', '__pydantic_fields_set__', '__pydantic_extra__', '__pydantic_private__'


def test_on_model_class():
    v = SchemaValidator(
        cs.model_schema(
            cls=MyModel,
            config=CoreConfig(str_max_length=5),
            schema=cs.model_fields_schema(fields={'f': cs.model_field(schema=cs.str_schema())}),
        )
    )
    assert 'max_length:Some(5)' in plain_repr(v)
    assert v.isinstance_python({'f': 'test'}) is True
    assert v.isinstance_python({'f': 'test long'}) is False


def test_field_priority_model():
    v = SchemaValidator(
        cs.model_schema(
            cls=MyModel,
            config=CoreConfig(str_max_length=10),
            schema=cs.model_fields_schema(fields={'f': cs.model_field(schema=cs.str_schema(max_length=5))}),
        )
    )
    assert 'max_length:Some(5)' in plain_repr(v)
    assert v.isinstance_python({'f': 'test'}) is True
    assert v.isinstance_python({'f': 'test long'}) is False


@pytest.mark.parametrize(
    'config,float_field_schema,input_value,expected',
    [
        (
            CoreConfig(),
            cs.float_schema(),
            {'x': 'nan'},
            IsInstance(MyModel) & HasAttributes(x=FunctionCheck(math.isnan)),
        ),
        (
            CoreConfig(allow_inf_nan=True),
            cs.float_schema(),
            {'x': 'nan'},
            IsInstance(MyModel) & HasAttributes(x=FunctionCheck(math.isnan)),
        ),
        (
            CoreConfig(allow_inf_nan=False),
            cs.float_schema(),
            {'x': 'nan'},
            Err('Input should be a finite number [type=finite_number,'),
        ),
        # field `allow_inf_nan` (if set) should have priority over global config
        (
            CoreConfig(allow_inf_nan=True),
            cs.float_schema(allow_inf_nan=False),
            {'x': 'nan'},
            Err('Input should be a finite number [type=finite_number,'),
        ),
        (
            CoreConfig(allow_inf_nan=False),
            cs.float_schema(allow_inf_nan=True),
            {'x': 'nan'},
            IsInstance(MyModel) & HasAttributes(x=FunctionCheck(math.isnan)),
        ),
    ],
    ids=repr,
)
def test_allow_inf_nan(config: CoreConfig, float_field_schema, input_value, expected):
    v = SchemaValidator(
        cs.model_schema(
            cls=MyModel,
            schema=cs.model_fields_schema(fields={'x': cs.model_field(schema=float_field_schema)}),
            config=config,
        )
    )
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            v.validate_python(input_value)
    else:
        output_dict = v.validate_python(input_value)
        assert output_dict == expected


@pytest.mark.parametrize(
    'config,input_str',
    (
        (CoreConfig(), 'type=string_type, input_value=123, input_type=int'),
        (CoreConfig(hide_input_in_errors=False), 'type=string_type, input_value=123, input_type=int'),
        (CoreConfig(hide_input_in_errors=True), 'type=string_type'),
    ),
)
def test_hide_input_in_errors(config, input_str):
    v = SchemaValidator(
        cs.model_schema(
            cls=MyModel, schema=cs.model_fields_schema(fields={'f': cs.model_field(schema=cs.str_schema())})
        ),
        config=config,
    )

    with pytest.raises(ValidationError, match=re.escape(f'Input should be a valid string [{input_str}]')):
        assert v.validate_python({'f': 123})


def test_cache_strings():
    v = SchemaValidator(cs.str_schema())
    assert 'cache_strings=True' in plain_repr(v)

    v = SchemaValidator(cs.str_schema(), config=CoreConfig(cache_strings=True))
    assert 'cache_strings=True' in plain_repr(v)

    v = SchemaValidator(cs.str_schema(), config=CoreConfig(cache_strings=False))
    assert 'cache_strings=False' in plain_repr(v)

    v = SchemaValidator(cs.str_schema(), config=CoreConfig(cache_strings='keys'))
    assert "cache_strings='keys'" in plain_repr(v)



import pytest

from pydantic_core import SchemaError, SchemaSerializer, core_schema


def test_custom_ser():
    s = SchemaSerializer(
        core_schema.definitions_schema(
            core_schema.list_schema(core_schema.definition_reference_schema('foobar')),
            [core_schema.int_schema(ref='foobar', serialization=core_schema.to_string_ser_schema(when_used='always'))],
        )
    )
    assert s.to_python([1, 2, 3]) == ['1', '2', '3']


def test_ignored_def():
    s = SchemaSerializer(
        core_schema.definitions_schema(
            core_schema.list_schema(core_schema.int_schema()),
            [core_schema.int_schema(ref='foobar', serialization=core_schema.to_string_ser_schema(when_used='always'))],
        )
    )
    assert s.to_python([1, 2, 3]) == [1, 2, 3]


def test_repeated_ref():
    with pytest.raises(SchemaError, match='SchemaError: Duplicate ref: `foobar`'):
        SchemaSerializer(
            core_schema.tuple_positional_schema(
                [
                    core_schema.definitions_schema(
                        core_schema.definition_reference_schema('foobar'), [core_schema.int_schema(ref='foobar')]
                    ),
                    core_schema.definitions_schema(
                        core_schema.definition_reference_schema('foobar'), [core_schema.int_schema(ref='foobar')]
                    ),
                ]
            )
        )


def test_repeat_after():
    with pytest.raises(SchemaError, match='SchemaError: Duplicate ref: `foobar`'):
        SchemaSerializer(
            core_schema.definitions_schema(
                core_schema.tuple_positional_schema(
                    [
                        core_schema.definitions_schema(
                            core_schema.definition_reference_schema('foobar'), [core_schema.int_schema(ref='foobar')]
                        ),
                        core_schema.definition_reference_schema('foobar'),
                    ]
                ),
                [core_schema.int_schema(ref='foobar')],
            )
        )


def test_deep():
    v = SchemaSerializer(
        core_schema.typed_dict_schema(
            {
                'a': core_schema.typed_dict_field(core_schema.int_schema()),
                'b': core_schema.typed_dict_field(
                    core_schema.definitions_schema(
                        core_schema.typed_dict_schema(
                            {
                                'c': core_schema.typed_dict_field(core_schema.int_schema()),
                                'd': core_schema.typed_dict_field(core_schema.definition_reference_schema('foobar')),
                            }
                        ),
                        [
                            core_schema.int_schema(
                                ref='foobar', serialization=core_schema.to_string_ser_schema(when_used='always')
                            )
                        ],
                    )
                ),
            }
        )
    )
    assert v.to_python({'a': 1, 'b': {'c': 2, 'd': 3}}) == {'a': 1, 'b': {'c': 2, 'd': '3'}}


def test_use_after():
    v = SchemaSerializer(
        core_schema.tuple_positional_schema(
            [
                core_schema.definitions_schema(
                    core_schema.definition_reference_schema('foobar'),
                    [
                        core_schema.int_schema(
                            ref='foobar', serialization=core_schema.to_string_ser_schema(when_used='always')
                        )
                    ],
                ),
                core_schema.definition_reference_schema('foobar'),
            ]
        )
    )
    assert v.to_python((1, 2)) == ('1', '2')


def test_defs_with_dict():
    s = SchemaSerializer(
        core_schema.definitions_schema(
            schema=core_schema.typed_dict_schema(
                {
                    'foo': core_schema.typed_dict_field(
                        core_schema.dict_schema(
                            keys_schema=core_schema.definition_reference_schema('key'),
                            values_schema=core_schema.definition_reference_schema('val'),
                        )
                    )
                }
            ),
            definitions=[core_schema.str_schema(ref='key'), core_schema.str_schema(ref='val')],
        )
    )

    assert s.to_json({'foo': {'key': 'val'}}) == b'{"foo":{"key":"val"}}'
    assert s.to_python({'foo': {'key': 'val'}}) == {'foo': {'key': 'val'}}



from decimal import Decimal

import pytest

from pydantic_core import SchemaError, SchemaValidator, ValidationError, core_schema
from pydantic_core import core_schema as cs

from ..conftest import PyAndJson


def test_chain():
    validator = SchemaValidator(
        cs.chain_schema(
            steps=[cs.str_schema(), core_schema.with_info_plain_validator_function(lambda v, info: Decimal(v))]
        )
    )

    assert validator.validate_python('1.44') == Decimal('1.44')
    assert validator.validate_python(b'1.44') == Decimal('1.44')


def test_chain_many():
    validator = SchemaValidator(
        cs.chain_schema(
            steps=[
                core_schema.with_info_plain_validator_function(lambda v, info: f'{v}-1'),
                core_schema.with_info_plain_validator_function(lambda v, info: f'{v}-2'),
                core_schema.with_info_plain_validator_function(lambda v, info: f'{v}-3'),
                core_schema.with_info_plain_validator_function(lambda v, info: f'{v}-4'),
            ]
        )
    )

    assert validator.validate_python('input') == 'input-1-2-3-4'


def test_chain_error():
    validator = SchemaValidator(cs.chain_schema(steps=[cs.str_schema(), cs.int_schema()]))

    assert validator.validate_python('123') == 123
    assert validator.validate_python(b'123') == 123

    with pytest.raises(ValidationError) as exc_info:
        validator.validate_python('abc')
    # insert_assert(exc_info.value.errors(include_url=False))
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'int_parsing',
            'loc': (),
            'msg': 'Input should be a valid integer, unable to parse string as an integer',
            'input': 'abc',
        }
    ]


@pytest.mark.parametrize(
    'input_value,expected', [('1.44', Decimal('1.44')), (1, Decimal(1)), (1.44, pytest.approx(1.44))]
)
def test_json(py_and_json: PyAndJson, input_value, expected):
    validator = py_and_json(
        {
            'type': 'chain',
            'steps': [
                {'type': 'union', 'choices': [{'type': 'str'}, {'type': 'float'}]},
                core_schema.with_info_plain_validator_function(lambda v, info: Decimal(v)),
            ],
        }
    )
    output = validator.validate_test(input_value)
    assert output == expected
    assert isinstance(output, Decimal)


def test_flatten():
    validator = SchemaValidator(
        cs.chain_schema(
            steps=[
                core_schema.with_info_plain_validator_function(lambda v, info: f'{v}-1'),
                cs.chain_schema(
                    steps=[
                        {
                            'type': 'function-plain',
                            'function': {'type': 'with-info', 'function': lambda v, info: f'{v}-2'},
                        },
                        {
                            'type': 'function-plain',
                            'function': {'type': 'with-info', 'function': lambda v, info: f'{v}-3'},
                        },
                    ]
                ),
            ]
        )
    )

    assert validator.validate_python('input') == 'input-1-2-3'
    assert validator.title == 'chain[function-plain[<lambda>()],function-plain[<lambda>()],function-plain[<lambda>()]]'


def test_chain_empty():
    with pytest.raises(SchemaError, match='One or more steps are required for a chain validator'):
        SchemaValidator(cs.chain_schema(steps=[]))


def test_chain_one():
    validator = SchemaValidator(
        cs.chain_schema(steps=[core_schema.with_info_plain_validator_function(lambda v, info: f'{v}-1')])
    )
    assert validator.validate_python('input') == 'input-1'
    assert validator.title == 'function-plain[<lambda>()]'



import pytest

from pydantic_core import ArgsKwargs, ValidationError
from pydantic_core import core_schema as cs

from ...conftest import PyAndJson


@pytest.mark.parametrize(
    ['input_value', 'expected'],
    (
        [ArgsKwargs(()), ((), {})],
        [ArgsKwargs((1, 2, 3)), ((1, 2, 3), {})],
        [{'args': ()}, ((), {})],
        [{'args': (1, 2, 3)}, ((1, 2, 3), {})],
        # Also validates against other sequence types, as long as it is
        # possible to validate it as a tuple:
        [{'args': [1, 2, 3]}, ((1, 2, 3), {})],
    ),
)
def test_var_args(py_and_json: PyAndJson, input_value, expected) -> None:
    """Test valid inputs against var-args parameters:

    ```python
    def func(*args: int):
        ...
    ```
    """
    v = py_and_json(
        cs.arguments_v3_schema(
            [
                cs.arguments_v3_parameter(name='args', schema=cs.int_schema(), mode='var_args'),
            ]
        )
    )

    assert v.validate_test(input_value) == expected


@pytest.mark.parametrize(
    ['input_value', 'err_loc'],
    (
        [ArgsKwargs(('not_an_int',)), (0,)],
        [
            ArgsKwargs(
                (
                    1,
                    'not_an_int',
                )
            ),
            (1,),
        ],
        [{'args': ['not_an_int']}, ('args', 0)],
        [{'args': [1, 'not_an_int']}, ('args', 1)],
    ),
)
def test_var_args_validation_error(py_and_json: PyAndJson, input_value, err_loc) -> None:
    """Test invalid inputs against var-args parameters:

    ```python
    def func(*args: int):
        ...

    func(1, 'not_an_int')
    ```
    """
    v = py_and_json(
        cs.arguments_v3_schema(
            [
                cs.arguments_v3_parameter(name='args', schema=cs.int_schema(), mode='var_args'),
            ]
        )
    )

    with pytest.raises(ValidationError) as exc_info:
        v.validate_test(input_value)

    error = exc_info.value.errors()[0]

    assert error['type'] == 'int_parsing'
    assert error['loc'] == err_loc


def test_var_args_invalid_tuple(py_and_json: PyAndJson) -> None:
    """Test invalid tuple-like input against var-args parameters in mapping validation mode."""
    v = py_and_json(
        cs.arguments_v3_schema(
            [
                cs.arguments_v3_parameter(name='args', schema=cs.int_schema(), mode='var_args'),
            ]
        )
    )

    with pytest.raises(ValidationError) as exc_info:
        v.validate_test({'args': 'not_a_tuple'})

    error = exc_info.value.errors()[0]

    assert error['type'] == 'tuple_type'
    assert error['loc'] == ('args',)



import json
import pickle
from datetime import timedelta

import pytest

from pydantic_core import core_schema
from pydantic_core._pydantic_core import SchemaSerializer


def repr_function(value, _info):
    return repr(value)


def test_basic_schema_serializer():
    s = SchemaSerializer(core_schema.dict_schema())
    s = pickle.loads(pickle.dumps(s))
    assert s.to_python({'a': 1, b'b': 2, 33: 3}) == {'a': 1, b'b': 2, 33: 3}
    assert s.to_python({'a': 1, b'b': 2, 33: 3, True: 4}, mode='json') == {'a': 1, 'b': 2, '33': 3, 'true': 4}
    assert s.to_json({'a': 1, b'b': 2, 33: 3, True: 4}) == b'{"a":1,"b":2,"33":3,"true":4}'

    assert s.to_python({(1, 2): 3}) == {(1, 2): 3}
    assert s.to_python({(1, 2): 3}, mode='json') == {'1,2': 3}
    assert s.to_json({(1, 2): 3}) == b'{"1,2":3}'


@pytest.mark.parametrize(
    'value,expected_python,expected_json',
    [(None, 'None', b'"None"'), (1, '1', b'"1"'), ([1, 2, 3], '[1, 2, 3]', b'"[1, 2, 3]"')],
)
def test_schema_serializer_capturing_function(value, expected_python, expected_json):
    # Test a SchemaSerializer that captures a function.
    s = SchemaSerializer(
        core_schema.any_schema(
            serialization=core_schema.plain_serializer_function_ser_schema(repr_function, info_arg=True)
        )
    )
    s = pickle.loads(pickle.dumps(s))
    assert s.to_python(value) == expected_python
    assert s.to_json(value) == expected_json
    assert s.to_python(value, mode='json') == json.loads(expected_json)


def test_schema_serializer_containing_config():
    s = SchemaSerializer(core_schema.timedelta_schema(), config={'ser_json_timedelta': 'float'})
    s = pickle.loads(pickle.dumps(s))

    assert s.to_python(timedelta(seconds=4, microseconds=500_000)) == timedelta(seconds=4, microseconds=500_000)
    assert s.to_python(timedelta(seconds=4, microseconds=500_000), mode='json') == 4.5
    assert s.to_json(timedelta(seconds=4, microseconds=500_000)) == b'4.5'


# Should be defined at the module level for pickling to work:
class Model:
    __pydantic_serializer__: SchemaSerializer
    __pydantic_complete__ = True


def test_schema_serializer_not_reused_when_unpickling() -> None:
    s = SchemaSerializer(
        core_schema.model_schema(
            cls=Model,
            schema=core_schema.model_fields_schema(fields={}, model_name='Model'),
            config={'title': 'Model'},
            ref='Model:123',
        )
    )

    Model.__pydantic_serializer__ = s
    assert 'Prebuilt' not in str(Model.__pydantic_serializer__)

    reconstructed = pickle.loads(pickle.dumps(Model.__pydantic_serializer__))
    assert 'Prebuilt' not in str(reconstructed)



from enum import Enum

from pydantic_core import SchemaSerializer, core_schema


def test_json_or_python():
    def s1(v: int) -> int:
        return v + 1

    def s2(v: int) -> int:
        return v + 2

    s = SchemaSerializer(
        core_schema.json_or_python_schema(
            core_schema.int_schema(serialization=core_schema.plain_serializer_function_ser_schema(s1)),
            core_schema.int_schema(serialization=core_schema.plain_serializer_function_ser_schema(s2)),
        )
    )

    assert s.to_json(0) == b'1'
    assert s.to_python(0) == 2


def test_json_or_python_enum_dict_key():
    # See https://github.com/pydantic/pydantic/issues/6795
    class MyEnum(str, Enum):
        A = 'A'
        B = 'B'

    print(MyEnum('A'))

    s = SchemaSerializer(
        core_schema.dict_schema(
            core_schema.json_or_python_schema(
                core_schema.str_schema(), core_schema.no_info_after_validator_function(MyEnum, core_schema.str_schema())
            ),
            core_schema.int_schema(),
        )
    )

    assert s.to_json({MyEnum.A: 1, MyEnum.B: 2}) == b'{"A":1,"B":2}'
    assert s.to_python({MyEnum.A: 1, MyEnum.B: 2}) == {MyEnum.A: 1, MyEnum.B: 2}



import copy
import pickle

import pytest
from typing_extensions import (  # noqa: UP035 (https://github.com/astral-sh/ruff/pull/18476)
    get_args,
    get_origin,
    get_type_hints,
)
from typing_inspection import typing_objects
from typing_inspection.introspection import UNKNOWN, AnnotationSource, inspect_annotation

from pydantic_core import CoreConfig, CoreSchema, CoreSchemaType, PydanticUndefined, core_schema
from pydantic_core._pydantic_core import (
    SchemaError,
    SchemaValidator,
    ValidationError,
    __version__,
    build_info,
    build_profile,
)


@pytest.mark.parametrize('obj', [ValidationError, SchemaValidator, SchemaError])
def test_module(obj):
    assert obj.__module__ == 'pydantic_core._pydantic_core'


def test_version():
    assert isinstance(__version__, str)
    assert '.' in __version__


def test_build_profile():
    assert build_profile in ('debug', 'release')


def test_build_info():
    assert isinstance(build_info, str)


def test_schema_error():
    err = SchemaError('test')
    assert isinstance(err, Exception)
    assert str(err) == 'test'
    assert repr(err) == 'SchemaError("test")'


def test_validation_error(pydantic_version):
    v = SchemaValidator(core_schema.int_schema())
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python(1.5)

    assert exc_info.value.title == 'int'
    assert exc_info.value.error_count() == 1
    assert (
        exc_info.value.errors(include_url=False)
        == exc_info.value.errors(include_url=False, include_context=False)
        == [
            {
                'type': 'int_from_float',
                'loc': (),
                'msg': 'Input should be a valid integer, got a number with a fractional part',
                'input': 1.5,
            }
        ]
    )
    # insert_assert(exc_info.value.errors())
    assert exc_info.value.errors() == [
        {
            'type': 'int_from_float',
            'loc': (),
            'msg': 'Input should be a valid integer, got a number with a fractional part',
            'input': 1.5,
            'url': f'https://errors.pydantic.dev/{pydantic_version}/v/int_from_float',
        }
    ]


def test_validation_error_include_context():
    v = SchemaValidator(core_schema.list_schema(max_length=2))
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python([1, 2, 3])

    assert exc_info.value.title == 'list[any]'
    assert exc_info.value.error_count() == 1
    # insert_assert(exc_info.value.errors(include_url=False))
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'too_long',
            'loc': (),
            'msg': 'List should have at most 2 items after validation, not 3',
            'input': [1, 2, 3],
            'ctx': {'field_type': 'List', 'max_length': 2, 'actual_length': 3},
        }
    ]
    # insert_assert(exc_info.value.errors(include_url=False, include_context=False))
    assert exc_info.value.errors(include_url=False, include_context=False) == [
        {
            'type': 'too_long',
            'loc': (),
            'msg': 'List should have at most 2 items after validation, not 3',
            'input': [1, 2, 3],
        }
    ]


def test_custom_title():
    v = SchemaValidator(core_schema.int_schema(), config=CoreConfig(title='MyInt'))
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python(1.5)

    assert exc_info.value.title == 'MyInt'


def test_validation_error_multiple(pydantic_version):
    class MyModel:
        # this is not required, but it avoids `__pydantic_fields_set__` being included in `__dict__`
        __slots__ = '__dict__', '__pydantic_fields_set__', '__pydantic_extra__', '__pydantic_private__'
        field_a: str
        field_b: int

    v = SchemaValidator(
        core_schema.model_schema(
            cls=MyModel,
            schema=core_schema.model_fields_schema(
                fields={
                    'x': core_schema.model_field(schema=core_schema.float_schema()),
                    'y': core_schema.model_field(schema=core_schema.int_schema()),
                }
            ),
        )
    )
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python({'x': 'x' * 60, 'y': 'y'})

    assert exc_info.value.title == 'MyModel'
    assert exc_info.value.error_count() == 2
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'float_parsing',
            'loc': ('x',),
            'msg': 'Input should be a valid number, unable to parse string as a number',
            'input': 'x' * 60,
        },
        {
            'type': 'int_parsing',
            'loc': ('y',),
            'msg': 'Input should be a valid integer, unable to parse string as an integer',
            'input': 'y',
        },
    ]
    assert repr(exc_info.value) == (
        '2 validation errors for MyModel\n'
        'x\n'
        '  Input should be a valid number, unable to parse string as a number '
        "[type=float_parsing, input_value='xxxxxxxxxxxxxxxxxxxxxxxx...xxxxxxxxxxxxxxxxxxxxxxx', input_type=str]\n"
        f'    For further information visit https://errors.pydantic.dev/{pydantic_version}/v/float_parsing\n'
        'y\n'
        '  Input should be a valid integer, unable to parse string as an integer '
        "[type=int_parsing, input_value='y', input_type=str]\n"
        f'    For further information visit https://errors.pydantic.dev/{pydantic_version}/v/int_parsing'
    )


def test_core_schema_type_literal():
    def get_type_value(schema_typeddict) -> str:
        annotation = get_type_hints(schema_typeddict, include_extras=True)['type']
        inspected_ann = inspect_annotation(annotation, annotation_source=AnnotationSource.TYPED_DICT)
        annotation = inspected_ann.type
        assert annotation is not UNKNOWN
        assert typing_objects.is_literal(get_origin(annotation)), (
            f"The 'type' key of core schemas must be a Literal form, got {get_origin(annotation)}"
        )
        args = get_args(annotation)
        assert len(args) == 1, (
            f"The 'type' key of core schemas must be a Literal form with a single element, got {len(args)} elements"
        )
        type_ = args[0]
        assert isinstance(type_, str), (
            f"The 'type' key of core schemas must be a Literal form with a single string element, got element of type {type(type_)}"
        )

        return type_

    schema_types = (get_type_value(x) for x in CoreSchema.__args__)
    schema_types = tuple(dict.fromkeys(schema_types))  # remove duplicates while preserving order
    if get_args(CoreSchemaType) != schema_types:
        literal = ''.join(f'\n    {e!r},' for e in schema_types)
        print(
            f'python code (near end of python/pydantic_core/core_schema.py):\n\nCoreSchemaType = Literal[{literal}\n]'
        )
        pytest.fail('core_schema.CoreSchemaType needs to be updated')


def test_undefined():
    with pytest.raises(NotImplementedError, match='UndefinedType'):
        PydanticUndefined.__class__()

    undefined_copy = copy.copy(PydanticUndefined)
    undefined_deepcopy = copy.deepcopy(PydanticUndefined)

    assert undefined_copy is PydanticUndefined
    assert undefined_deepcopy is PydanticUndefined

    assert pickle.loads(pickle.dumps(PydanticUndefined)) is PydanticUndefined


def test_unicode_error_input_repr() -> None:
    """https://github.com/pydantic/pydantic/issues/6448"""

    validator = SchemaValidator(core_schema.int_schema())

    danger_str = 'ÿ' * 1000
    expected = "1 validation error for int\n  Input should be a valid integer, unable to parse string as an integer [type=int_parsing, input_value='ÿÿÿÿÿÿÿÿÿÿÿÿ...ÿÿÿÿÿÿÿÿÿÿÿ', input_type=str]"
    with pytest.raises(ValidationError) as exc_info:
        validator.validate_python(danger_str)
    actual = repr(exc_info.value).split('For further information visit ')[0].strip()

    assert expected == actual


def test_core_schema_import_field_validation_info():
    with pytest.warns(DeprecationWarning, match='`FieldValidationInfo` is deprecated, use `ValidationInfo` instead.'):
        core_schema.FieldValidationInfo


def test_core_schema_import_missing():
    with pytest.raises(AttributeError, match="module 'pydantic_core' has no attribute 'foobar'"):
        core_schema.foobar



import re
import sys
from collections import OrderedDict
from collections.abc import Mapping
from typing import Any

import pytest
from dirty_equals import HasRepr, IsStr

from pydantic_core import SchemaValidator, ValidationError
from pydantic_core import core_schema as cs

from ..conftest import Err, PyAndJson


def test_dict(py_and_json: PyAndJson):
    v = py_and_json({'type': 'dict', 'keys_schema': {'type': 'int'}, 'values_schema': {'type': 'int'}})
    assert v.validate_test({'1': 2, '3': 4}) == {1: 2, 3: 4}
    v = py_and_json({'type': 'dict', 'strict': True, 'keys_schema': {'type': 'int'}, 'values_schema': {'type': 'int'}})
    assert v.validate_test({'1': 2, '3': 4}) == {1: 2, 3: 4}
    assert v.validate_test({}) == {}
    with pytest.raises(ValidationError, match=re.escape('[type=dict_type, input_value=[], input_type=list]')):
        v.validate_test([])


@pytest.mark.parametrize(
    'input_value,expected',
    [
        ({'1': b'1', '2': b'2'}, {'1': '1', '2': '2'}),
        (OrderedDict(a=b'1', b='2'), {'a': '1', 'b': '2'}),
        ({}, {}),
        ('foobar', Err("Input should be a valid dictionary [type=dict_type, input_value='foobar', input_type=str]")),
        ([], Err('Input should be a valid dictionary [type=dict_type,')),
        ([('x', 'y')], Err('Input should be a valid dictionary [type=dict_type,')),
        ([('x', 'y'), ('z', 'z')], Err('Input should be a valid dictionary [type=dict_type,')),
        ((), Err('Input should be a valid dictionary [type=dict_type,')),
        ((('x', 'y'),), Err('Input should be a valid dictionary [type=dict_type,')),
        ((type('Foobar', (), {'x': 1})()), Err('Input should be a valid dictionary [type=dict_type,')),
    ],
    ids=repr,
)
def test_dict_cases(input_value, expected):
    v = SchemaValidator(cs.dict_schema(keys_schema=cs.str_schema(), values_schema=cs.str_schema()))
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            v.validate_python(input_value)
    else:
        assert v.validate_python(input_value) == expected


def test_dict_value_error(py_and_json: PyAndJson):
    v = py_and_json({'type': 'dict', 'values_schema': {'type': 'int'}})
    assert v.validate_test({'a': 2, 'b': '4'}) == {'a': 2, 'b': 4}
    with pytest.raises(ValidationError, match='Input should be a valid integer') as exc_info:
        v.validate_test({'a': 2, 'b': 'wrong'})
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'int_parsing',
            'loc': ('b',),
            'msg': 'Input should be a valid integer, unable to parse string as an integer',
            'input': 'wrong',
        }
    ]


def test_dict_error_key_int():
    v = SchemaValidator(cs.dict_schema(values_schema=cs.int_schema()))
    with pytest.raises(ValidationError, match='Input should be a valid integer') as exc_info:
        v.validate_python({1: 2, 3: 'wrong', -4: 'wrong2'})
    # insert_assert(exc_info.value.errors(include_url=False))
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'int_parsing',
            'loc': (3,),
            'msg': 'Input should be a valid integer, unable to parse string as an integer',
            'input': 'wrong',
        },
        {
            'type': 'int_parsing',
            'loc': (-4,),
            'msg': 'Input should be a valid integer, unable to parse string as an integer',
            'input': 'wrong2',
        },
    ]


def test_dict_error_key_other():
    v = SchemaValidator(cs.dict_schema(values_schema=cs.int_schema()))
    with pytest.raises(ValidationError, match='Input should be a valid integer') as exc_info:
        v.validate_python({1: 2, (1, 2): 'wrong'})
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'int_parsing',
            'loc': ('(1, 2)',),
            'msg': 'Input should be a valid integer, unable to parse string as an integer',
            'input': 'wrong',
        }
    ]


def test_dict_any_value():
    v = SchemaValidator(cs.dict_schema(keys_schema=cs.str_schema()))
    v = SchemaValidator(cs.dict_schema(keys_schema=cs.str_schema()))
    assert v.validate_python({'1': 1, '2': 'a', '3': None}) == {'1': 1, '2': 'a', '3': None}


def test_mapping():
    class MyMapping(Mapping):
        def __init__(self, d):
            self._d = d

        def __getitem__(self, key):
            return self._d[key]

        def __iter__(self):
            return iter(self._d)

        def __len__(self):
            return len(self._d)

    v = SchemaValidator(cs.dict_schema(keys_schema=cs.int_schema(), values_schema=cs.int_schema()))
    assert v.validate_python(MyMapping({'1': 2, 3: '4'})) == {1: 2, 3: 4}
    v = SchemaValidator(cs.dict_schema(strict=True, keys_schema=cs.int_schema(), values_schema=cs.int_schema()))
    with pytest.raises(ValidationError, match='Input should be a valid dictionary'):
        v.validate_python(MyMapping({'1': 2, 3: '4'}))


def test_key_error():
    v = SchemaValidator(cs.dict_schema(keys_schema=cs.int_schema(), values_schema=cs.int_schema()))
    assert v.validate_python({'1': True}) == {1: 1}
    with pytest.raises(ValidationError, match=re.escape('x.[key]\n  Input should be a valid integer')) as exc_info:
        v.validate_python({'x': 1})
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'int_parsing',
            'loc': ('x', '[key]'),
            'msg': 'Input should be a valid integer, unable to parse string as an integer',
            'input': 'x',
        }
    ]


def test_mapping_error():
    class BadMapping(Mapping):
        def __getitem__(self, key):
            raise None

        def __iter__(self):
            raise RuntimeError('intentional error')

        def __len__(self):
            return 1

    v = SchemaValidator(cs.dict_schema(keys_schema=cs.int_schema(), values_schema=cs.int_schema()))
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python(BadMapping())

    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'mapping_type',
            'loc': (),
            'msg': 'Input should be a valid mapping, error: RuntimeError: intentional error',
            'input': HasRepr(IsStr(regex='.+BadMapping object at.+')),
            'ctx': {'error': 'RuntimeError: intentional error'},
        }
    ]


@pytest.mark.parametrize('mapping_items', [[(1,)], ['foobar'], [(1, 2, 3)], 'not list'])
def test_mapping_error_yield_1(mapping_items):
    class BadMapping(Mapping):
        def items(self):
            return mapping_items

        def __iter__(self):
            pytest.fail('unexpected call to __iter__')

        def __getitem__(self, key):
            pytest.fail('unexpected call to __getitem__')

        def __len__(self):
            return 1

    v = SchemaValidator(cs.dict_schema(keys_schema=cs.int_schema(), values_schema=cs.int_schema()))
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python(BadMapping())

    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'mapping_type',
            'loc': (),
            'msg': 'Input should be a valid mapping, error: Mapping items must be tuples of (key, value) pairs',
            'input': HasRepr(IsStr(regex='.+BadMapping object at.+')),
            'ctx': {'error': 'Mapping items must be tuples of (key, value) pairs'},
        }
    ]


@pytest.mark.parametrize(
    'kwargs,input_value,expected',
    [
        ({}, {'1': 1, '2': 2}, {'1': 1, '2': 2}),
        (
            {'min_length': 3},
            {'1': 1, '2': 2, '3': 3.0, '4': [1, 2, 3, 4]},
            {'1': 1, '2': 2, '3': 3.0, '4': [1, 2, 3, 4]},
        ),
        (
            {'min_length': 3},
            {1: '2', 3: '4'},
            Err('Dictionary should have at least 3 items after validation, not 2 [type=too_short,'),
        ),
        ({'max_length': 4}, {'1': 1, '2': 2, '3': 3.0}, {'1': 1, '2': 2, '3': 3.0}),
        (
            {'max_length': 3},
            {'1': 1, '2': 2, '3': 3.0, '4': [1, 2, 3, 4]},
            Err('Dictionary should have at most 3 items after validation, not 4 [type=too_long,'),
        ),
    ],
)
def test_dict_length_constraints(kwargs: dict[str, Any], input_value, expected):
    v = SchemaValidator(cs.dict_schema(**kwargs))
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            v.validate_python(input_value)
    else:
        assert v.validate_python(input_value) == expected


def test_json_dict():
    v = SchemaValidator(cs.dict_schema(keys_schema=cs.int_schema(), values_schema=cs.int_schema()))
    assert v.validate_json('{"1": 2, "3": 4}') == {1: 2, 3: 4}
    with pytest.raises(ValidationError) as exc_info:
        v.validate_json('1')
    assert exc_info.value.errors(include_url=False) == [
        {'type': 'dict_type', 'loc': (), 'msg': 'Input should be an object', 'input': 1}
    ]


def test_dict_complex_key():
    v = SchemaValidator(cs.dict_schema(keys_schema=cs.complex_schema(strict=True), values_schema=cs.str_schema()))
    assert v.validate_python({complex(1, 2): '1'}) == {complex(1, 2): '1'}
    with pytest.raises(ValidationError, match='Input should be an instance of complex'):
        assert v.validate_python({'1+2j': b'1'}) == {complex(1, 2): '1'}

    v = SchemaValidator(cs.dict_schema(keys_schema=cs.complex_schema(), values_schema=cs.str_schema()))
    with pytest.raises(
        ValidationError, match='Input should be a valid python complex object, a number, or a valid complex string'
    ):
        v.validate_python({'1+2ja': b'1'})


def test_json_dict_complex_key():
    v = SchemaValidator(cs.dict_schema(keys_schema=cs.complex_schema(), values_schema=cs.int_schema()))
    assert v.validate_json('{"1+2j": 2, "-3": 4}') == {complex(1, 2): 2, complex(-3, 0): 4}
    assert v.validate_json('{"1+2j": 2, "infj": 4}') == {complex(1, 2): 2, complex(0, float('inf')): 4}
    with pytest.raises(ValidationError, match='Input should be a valid complex string'):
        v.validate_json('{"1+2j": 2, "": 4}') == {complex(1, 2): 2, complex(0, float('inf')): 4}


@pytest.mark.parametrize(
    ('fail_fast', 'expected'),
    [
        pytest.param(
            True,
            [
                {
                    'type': 'int_parsing',
                    'loc': ('a', '[key]'),
                    'msg': 'Input should be a valid integer, unable to parse string as an integer',
                    'input': 'a',
                },
            ],
            id='fail_fast',
        ),
        pytest.param(
            False,
            [
                {
                    'type': 'int_parsing',
                    'loc': ('a', '[key]'),
                    'msg': 'Input should be a valid integer, unable to parse string as an integer',
                    'input': 'a',
                },
                {
                    'type': 'int_parsing',
                    'loc': ('a',),
                    'msg': 'Input should be a valid integer, unable to parse string as an integer',
                    'input': 'b',
                },
                {
                    'type': 'int_parsing',
                    'loc': ('c', '[key]'),
                    'msg': 'Input should be a valid integer, unable to parse string as an integer',
                    'input': 'c',
                },
                {
                    'type': 'int_parsing',
                    'loc': ('c',),
                    'msg': 'Input should be a valid integer, unable to parse string as an integer',
                    'input': 'd',
                },
            ],
            id='not_fail_fast',
        ),
    ],
)
def test_dict_fail_fast(fail_fast, expected):
    v = SchemaValidator(
        {'type': 'dict', 'keys_schema': {'type': 'int'}, 'values_schema': {'type': 'int'}, 'fail_fast': fail_fast}
    )

    with pytest.raises(ValidationError) as exc_info:
        v.validate_python({'a': 'b', 'c': 'd'})

    assert exc_info.value.errors(include_url=False) == expected


@pytest.mark.skipif(
    sys.implementation.name == 'graalpy',
    reason='GraalPy has a bug where PyMapping.items() does not preserve OrderedDict order. See: https://github.com/oracle/graalpython/issues/553',
)
@pytest.mark.parametrize('strict', [True, False])
def test_ordered_dict_key_order_preservation(strict):
    # GH 12273
    v = SchemaValidator(cs.dict_schema(keys_schema=cs.str_schema(), values_schema=cs.int_schema()))

    # Original issue
    foo = OrderedDict({'a': 1, 'b': 2})
    foo.move_to_end('a')

    result = v.validate_python(foo, strict=strict)
    assert list(result.keys()) == list(foo.keys()) == ['b', 'a']
    assert result == {'b': 2, 'a': 1}

    # More complex case
    foo2 = OrderedDict({'x': 1, 'y': 2, 'z': 3})
    foo2.move_to_end('x')

    result2 = v.validate_python(foo2, strict=strict)
    assert list(result2.keys()) == list(foo2.keys()) == ['y', 'z', 'x']
    assert result2 == {'y': 2, 'z': 3, 'x': 1}



import pytest

from pydantic_core import ArgsKwargs, ValidationError
from pydantic_core import core_schema as cs

from ...conftest import PyAndJson


@pytest.mark.parametrize(
    'input_value',
    [
        ArgsKwargs((), {'a': 1, 'b': True}),
        ArgsKwargs((), {'a': 1}),
        {'a': 1, 'b': True},
        {'a': 1},
    ],
)
def test_keyword_only(py_and_json: PyAndJson, input_value) -> None:
    """Test valid inputs against keyword-only parameters:

    ```python
    def func(*, a: int, b: bool = True):
        ...
    ```
    """
    v = py_and_json(
        cs.arguments_v3_schema(
            [
                cs.arguments_v3_parameter(name='a', schema=cs.int_schema(), mode='keyword_only'),
                cs.arguments_v3_parameter(
                    name='b', schema=cs.with_default_schema(cs.bool_schema(), default=True), mode='keyword_only'
                ),
            ]
        )
    )

    assert v.validate_test(input_value) == ((), {'a': 1, 'b': True})


@pytest.mark.parametrize(
    'input_value',
    [ArgsKwargs((), {'a': 'not_an_int'}), {'a': 'not_an_int'}],
)
def test_keyword_only_validation_error(py_and_json: PyAndJson, input_value) -> None:
    """Test invalid inputs against keyword-only parameters:

    ```python
    def func(*, a: int):
        ...

    func('not_an_int')
    ```
    """
    v = py_and_json(
        cs.arguments_v3_schema(
            [
                cs.arguments_v3_parameter(name='a', schema=cs.int_schema(), mode='keyword_only'),
            ]
        )
    )

    with pytest.raises(ValidationError) as exc_info:
        v.validate_test(input_value)

    error = exc_info.value.errors()[0]

    assert error['type'] == 'int_parsing'
    assert error['loc'] == ('a',)


@pytest.mark.parametrize(
    'input_value',
    [ArgsKwargs((), {}), {}],
)
def test_keyword_only_error_required(py_and_json: PyAndJson, input_value) -> None:
    """Test missing inputs against keyword-only parameters:

    ```python
    def func(*, a: int):
        ...

    func()
    ```
    """
    v = py_and_json(
        cs.arguments_v3_schema(
            [
                cs.arguments_v3_parameter(name='a', schema=cs.int_schema(), mode='keyword_only'),
            ]
        )
    )

    with pytest.raises(ValidationError) as exc_info:
        v.validate_test(input_value)

    error = exc_info.value.errors()[0]

    assert error['type'] == 'missing_keyword_only_argument'
    assert error['loc'] == ('a',)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from pydantic_core import core_schema as cs

N = 5  # arbitrary number that takes ~0.05s per run


class MyModel:
    # __slots__ is not required, but it avoids __pydantic_fields_set__ falling into __dict__
    __slots__ = '__dict__', '__pydantic_fields_set__', '__pydantic_extra__', '__pydantic_private__'


def schema_using_defs() -> cs.CoreSchema:
    definitions: list[cs.CoreSchema] = [
        {'type': 'int', 'ref': 'int'},
        {
            'type': 'model',
            'cls': MyModel,
            'schema': {
                'type': 'model-fields',
                'fields': {
                    str(c): {'type': 'model-field', 'schema': {'type': 'definition-ref', 'schema_ref': 'int'}}
                    for c in range(N)
                },
            },
            'ref': f'model_{N}',
        },
    ]
    level = N
    for level in reversed(range(N)):
        definitions.append(
            {
                'type': 'model',
                'cls': MyModel,
                'schema': {
                    'type': 'model-fields',
                    'fields': {
                        str(c): {
                            'type': 'model-field',
                            'schema': {'type': 'definition-ref', 'schema_ref': f'model_{level + 1}'},
                        }
                        for c in range(N)
                    },
                },
                'ref': f'model_{level}',
            }
        )
    return {
        'type': 'definitions',
        'definitions': definitions,
        'schema': {'type': 'definition-ref', 'schema_ref': 'model_0'},
    }


def inlined_schema() -> cs.CoreSchema:
    level = N
    schema: cs.CoreSchema = {
        'type': 'model',
        'cls': MyModel,
        'schema': {
            'type': 'model-fields',
            'fields': {str(c): {'type': 'model-field', 'schema': {'type': 'int'}} for c in range(N)},
        },
        'ref': f'model_{N}',
    }
    for level in reversed(range(N)):
        schema = {
            'type': 'model',
            'cls': MyModel,
            'schema': {
                'type': 'model-fields',
                'fields': {str(c): {'type': 'model-field', 'schema': schema} for c in range(N)},
            },
            'ref': f'model_{level}',
        }
    return schema


def input_data_valid(levels: int = N) -> Any:
    data = {str(c): 1 for c in range(N)}
    for _ in range(levels):
        data = {str(c): data for c in range(N)}
    return data


if __name__ == '__main__':
    from pydantic_core import SchemaValidator

    SchemaValidator(schema_using_defs()).validate_python(input_data_valid())
    SchemaValidator(inlined_schema()).validate_python(input_data_valid())
